# sage_setup: distribution = sagemath-schemes
# sage.doctest: needs sage.rings.finite_rings sage.rings.number_field
r"""
Saturation of Mordell-Weil groups of elliptic curves over number fields

Points `P_1`, `\dots`, `P_r` in `E(K)`, where `E` is an elliptic curve
over a number field `K`, are said to be `p`-saturated if no linear
combination `\sum n_iP_i` is divisible by `p` in `E(K)` except
trivially when all `n_i` are multiples of `p`.  The points are said to
be saturated if they are `p`-saturated at all primes; this is always
true for all but finitely many primes since `E(K)` is a
finitely-generated Abelian group.

The process of `p`-saturating a given set of points is implemented
here.  The naive algorithm simply checks all `(p^r-1)/(p-1)`
projective combinations of the points, testing each to see if it can
be divided by `p`.  If this occurs then we replace one of the points
and continue.  The function :meth:`p_saturation` does one step of
this, while :meth:`full_p_saturation` repeats until the points are
`p`-saturated.  A more sophisticated algorithm for `p`-saturation is
implemented which is much more efficient for large `p` and `r`, and
involves computing the reduction of the points modulo auxiliary primes
to obtain linear conditions modulo `p` which must be satisfied by the
coefficients `a_i` of any nontrivial relation.  When the points are
already `p`-saturated this sieving technique can prove their
saturation quickly.

The method :meth:`saturation` of the class :class:`EllipticCurve_number_field`
applies full `p`-saturation at any given set of primes, or can compute
a bound on the primes `p` at which the given points may not be
`p`-saturated.  This involves computing a lower bound for the
canonical height of points of infinite order, together with estimates
from the geometry of numbers.

AUTHORS:

- Robert Bradshaw

- John Cremona
"""
#*****************************************************************************
#       Copyright (C) 2017 Robert Bradshaw <robertwb@math.washington.edu>
#                          John Cremona <john.cremona@gmail.com>
#                          William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.rings.finite_rings.finite_field_constructor import GF
from sage.rings.integer_ring import ZZ
from sage.arith.misc import kronecker as kro
from sage.structure.sage_object import SageObject


def reduce_mod_q(x, amodq):
    r"""The reduction of ``x`` modulo the prime ideal defined by ``amodq``.

    INPUT:

    - ``x`` -- an element of a  number field `K`

    - ``amodq`` -- an element of `GF(q)` which is a root mod `q` of
      the defining polynomial of `K`.  This defines a degree 1 prime
      ideal `Q=(q,\alpha-a)` of `K=\QQ(\alpha)`, where `a \bmod q` =
      ``amodq``.

    OUTPUT: the image of ``x`` in the residue field of `K` at the prime `Q`

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.saturation import reduce_mod_q
        sage: x = polygen(QQ)
        sage: pol = x^3 - x^2 - 3*x + 1
        sage: K.<a> = NumberField(pol)
        sage: [(q, [(amodq, reduce_mod_q(1 - a + a^4, amodq))
        ....:       for amodq in sorted(pol.roots(GF(q), multiplicities=False))])
        ....:  for q in primes(50, 70)]
        [(53, []),
         (59, [(36, 28)]),
         (61, [(40, 35)]),
         (67, [(10, 8), (62, 28), (63, 60)])]
    """
    Fq = amodq.parent()
    try:
        return x.lift().change_ring(Fq)(amodq)
    except AttributeError: # in case x is in QQ
        return Fq(x)


class EllipticCurveSaturator(SageObject):
    r"""
    Class for saturating points on an elliptic curve over a number field.

    INPUT:

    - ``E`` -- an elliptic curve defined over a number field, or `\QQ`

    - ``verbose`` -- boolean (default: ``False``); verbosity flag

    .. NOTE::

        This function is not normally called directly by users, who
        may access the data via methods of the EllipticCurve
        classes.
    """
    def __init__(self, E, verbose=False):
        r"""
        Initialize the saturator.

        INPUT:

        - ``E`` -- an elliptic curve defined over a number field

        - ``verbose`` -- boolean (default: ``False``); verbosity flag
        """
        self._verbose = verbose
        self._curve = E
        self._N = E.discriminant().norm()
        self._field = K = E.base_field()
        if K.absolute_degree() == 1:
            from sage.rings.rational_field import QQ
            from sage.rings.polynomial.polynomial_ring import polygen
            self._Kpol = polygen(QQ)
        else:
            self._Kpol = K.defining_polynomial()
        self._D = self._Kpol.discriminant()
        self._reductions = {}
        self._lincombs = {}
        self._torsion_gens = [t.element() for t in E.torsion_subgroup().gens()]
        self._reductions = {}
        # This will hold a dictionary with keys (q,aq) with q prime
        # and aq a root of K's defining polynomial mod q, and values
        # (n,gens) where n is the cardinality of the reduction of E
        # and gens are generators of that reduction.

    def add_reductions(self, q):
        r"""
        Add reduction data at primes above ``q`` if not already there.

        INPUT:

        - ``q`` -- a prime number not dividing the defining polynomial
          of ``self.__field``

        OUTPUT:

        Returns nothing, but updates self._reductions dictionary for
        key ``q`` to a dict whose keys are the roots of the defining
        polynomial mod ``q`` and values tuples (``nq``, ``Eq``) where
        ``Eq`` is an elliptic curve over `GF(q)` and ``nq`` its
        cardinality.  If ``q`` divides the conductor norm or order
        discriminant nothing is added.

        EXAMPLES:

        Over `\QQ`::

            sage: from sage.schemes.elliptic_curves.saturation import EllipticCurveSaturator

            sage: # needs database_cremona_mini_ellcurve
            sage: E = EllipticCurve('11a1')
            sage: saturator = EllipticCurveSaturator(E)
            sage: saturator._reductions
            {}
            sage: saturator.add_reductions(19)
            sage: saturator._reductions
            {19: {0: (20, Elliptic Curve defined by y^2 + y = x^3 + 18*x^2 + 9*x + 18
                           over Finite Field of size 19)}}

        Over a number field::

            sage: x = polygen(QQ);  K.<a> = NumberField(x^2 + 2)
            sage: E = EllipticCurve(K, [0,1,0,a,a])
            sage: from sage.schemes.elliptic_curves.saturation import EllipticCurveSaturator
            sage: saturator = EllipticCurveSaturator(E)
            sage: for q in primes(20):
            ....:     saturator.add_reductions(q)
            sage: saturator._reductions
            {2:  {},
             3:  {},
             5:  {},
             7:  {},
             11: {3:  (16, Elliptic Curve defined by y^2 = x^3 + x^2 + 3*x + 3
                            over Finite Field of size 11),
                  8:  (8,  Elliptic Curve defined by y^2 = x^3 + x^2 + 8*x + 8
                            over Finite Field of size 11)},
             13: {},
             17: {7:  (20, Elliptic Curve defined by y^2 = x^3 + x^2 + 7*x + 7
                            over Finite Field of size 17),
                  10: (18, Elliptic Curve defined by y^2 = x^3 + x^2 + 10*x + 10
                            over Finite Field of size 17)},
             19: {6:  (16, Elliptic Curve defined by y^2 = x^3 + x^2 + 6*x + 6
                            over Finite Field of size 19),
                  13: (12, Elliptic Curve defined by y^2 = x^3 + x^2 + 13*x + 13
                            over Finite Field of size 19)}}
        """
        if q in self._reductions:
            return
        self._reductions[q] = redmodq = {}
        if q.divides(self._N) or q.divides(self._D):
            return
        from sage.schemes.elliptic_curves.constructor import EllipticCurve
        for amodq in sorted(self._Kpol.roots(GF(q), multiplicities=False)):
            Eq = EllipticCurve([reduce_mod_q(ai, amodq) for ai in self._curve.ainvs()])
            nq = Eq.cardinality()
            redmodq[amodq] = (nq, Eq)

    def full_p_saturation(self, Plist, p):
        r"""Full `p`-saturation of ``Plist``.

        INPUT:

        - ``Plist`` -- list of independent points on one elliptic curve

        - ``p`` -- integer; a prime number

        OUTPUT:

        (``newPlist``, ``exponent``) where ``newPlist`` has the same
        length as ``Plist`` and spans the `p`-saturation of the span
        of ``Plist``, which contains that span with index
        ``p**exponent``.

        EXAMPLES::

            sage: # needs database_cremona_mini_ellcurve
            sage: from sage.schemes.elliptic_curves.saturation import EllipticCurveSaturator
            sage: E = EllipticCurve('389a')
            sage: K.<i> = QuadraticField(-1)
            sage: EK = E.change_ring(K)
            sage: P = EK(1 + i, -1 - 2*i)
            sage: saturator = EllipticCurveSaturator(EK, verbose=True)
            sage: saturator.full_p_saturation([8*P], 2)
             --starting full 2-saturation
            Points were not 2-saturated, exponent was 3
            ([(i + 1 : -2*i - 1 : 1)], 3)
            sage: Q = EK(0, 0)
            sage: R = EK(-1, 1)
            sage: saturator = EllipticCurveSaturator(EK, verbose=False)
            sage: saturator.full_p_saturation([P, Q, R], 3)
            ([(i + 1 : -2*i - 1 : 1), (0 : 0 : 1), (-1 : 1 : 1)], 0)

        An example where the points are not 7-saturated and we gain
        index exponent 1.  Running this example with verbose=True
        would show that it uses the code for when the reduction has
        `p`-rank 2 (which occurs for the reduction modulo `(16-5i)`),
        which uses the Weil pairing::

            sage: saturator.full_p_saturation([P, Q + 3*R, Q - 4*R], 7)                 # needs database_cremona_mini_ellcurve
            ([(i + 1 : -2*i - 1 : 1),
              (2869/676 : 154413/17576 : 1),
              (-7095/502681 : -366258864/356400829 : 1)], 1)
        """
        if not Plist:
            return Plist, ZZ.zero()

        exponent = ZZ(0)

        # To handle p-torsion, we must add any torsion generators whose
        # order is divisible by p to the list of points.  Note that it is
        # not correct to add generators of the p-torsion here, we actually
        # need generators of p-cotorsion.  If there are any of these, we
        # cannot use the supplied dict of linear combinations, so we reset
        # this.  The torsion points are removed before returning the
        # saturated list.

        verbose = self._verbose
        if verbose:
            print(" --starting full %s-saturation" % p)

        n = len(Plist)  # number of points supplied & to be returned
        Plist = Plist + [T for T in self._torsion_gens if p.divides(T.order())]
        nx = len(Plist) # number of points including relevant torsion
        extra_torsion = nx-n
        if extra_torsion:
            if verbose:
                print("Adding {} torsion generators before {}-saturation".format(extra_torsion,p))

        res = self.p_saturation(Plist, p)
        while res: # res is either False or (i, newP)
            exponent += 1
            Plist[res[0]] = res[1]
            res = self.p_saturation(Plist, p)

        if extra_torsion:
            # remove the torsion points
            if verbose:
                print("Removing the torsion generators after %s-saturation" % p)
            Plist = Plist[:n]

        if verbose:
            if exponent:
                print("Points were not %s-saturated, exponent was %s" % (p,exponent))
            else:
                print("Points were %s-saturated" % p)

        return (Plist, exponent)

    def p_saturation(self, Plist, p, sieve=True):
        r"""Checks whether the list of points is `p`-saturated.

        INPUT:

        - ``Plist`` -- list of independent points on one elliptic curve

        - ``p`` -- integer; a prime number

        - ``sieve`` -- boolean; if ``True``, use a sieve (when there are at
          least 2 points), otherwise test all combinations

        .. NOTE::

            The sieve is much more efficient when the points are
            saturated and the number of points or the prime are large.

        OUTPUT:

        Either ``False`` if the points are `p`-saturated, or (``i``,
        ``newP``) if they are not `p`-saturated, in which case after
        replacing the i'th point with ``newP``, the subgroup generated
        contains that generated by ``Plist`` with index `p`.

        EXAMPLES::

            sage: # needs database_cremona_mini_ellcurve
            sage: from sage.schemes.elliptic_curves.saturation import EllipticCurveSaturator
            sage: E = EllipticCurve('389a')
            sage: K.<i> = QuadraticField(-1)
            sage: EK = E.change_ring(K)
            sage: P = EK(1 + i, -1 - 2*i)
            sage: saturator = EllipticCurveSaturator(EK)
            sage: saturator.p_saturation([P], 2)
            False
            sage: saturator.p_saturation([2*P], 2)
            (0, (i + 1 : -2*i - 1 : 1))
            sage: Q = EK(0, 0)
            sage: R = EK(-1, 1)
            sage: saturator.p_saturation([P, Q, R], 3)
            False

        Here we see an example where 19-saturation is proved, with the
        verbose flag set to ``True`` so that we can see what is going on::

            sage: # needs database_cremona_mini_ellcurve
            sage: saturator = EllipticCurveSaturator(EK, verbose=True)
            sage: saturator.p_saturation([P, Q, R], 19)
            Using sieve method to saturate...
            E has 19-torsion over Finite Field of size 197, projecting points
            --> [(15 : 168 : 1), (0 : 0 : 1), (196 : 1 : 1)]
            --rank is now 1
            E has 19-torsion over Finite Field of size 197, projecting points
            --> [(184 : 27 : 1), (0 : 0 : 1), (196 : 1 : 1)]
            --rank is now 2
            E has 19-torsion over Finite Field of size 293, projecting points
            --> [(139 : 16 : 1), (0 : 0 : 1), (292 : 1 : 1)]
             --rank is now 3
            Reached full rank: points were 19-saturated
            False

        An example where the points are not 11-saturated::

            sage: # needs database_cremona_mini_ellcurve
            sage: saturator = EllipticCurveSaturator(EK, verbose=False)
            sage: res = saturator.p_saturation([P + 5*Q, P - 6*Q, R], 11); res
            (0, (-5783311/14600041*i + 1396143/14600041
                 : 37679338314/55786756661*i + 3813624227/55786756661 : 1))

        That means that the 0'th point may be replaced by the displayed
        point to achieve an index gain of 11::

            sage: saturator.p_saturation([res[1], P - 6*Q, R], 11)                      # needs database_cremona_mini_ellcurve
            False

        TESTS:

        See :issue:`27387`::

            sage: x = polygen(QQ)
            sage: K.<a> = NumberField(x^2 - x - 26)
            sage: E = EllipticCurve([a, 1 - a, 0, 93 - 16*a, 3150 - 560*a])
            sage: P = E([65 - 35*a/3, (959*a-5377)/9])
            sage: T = E.torsion_points()[0]
            sage: from sage.schemes.elliptic_curves.saturation import EllipticCurveSaturator
            sage: saturator = EllipticCurveSaturator(E, True)
            sage: saturator.p_saturation([P, T], 2)
            Using sieve method to saturate...
            ...
            -- points were not 2-saturated, gaining index 2
            (1, (0 : 1 : 0))

        A CM example where large sieving primes are needed (LMFDB
        label 2.0.3.1-50625.1-CMb2)::

            sage: x = polygen(QQ)
            sage: K.<a> = NumberField(x^2 - x + 1)
            sage: E = EllipticCurve(K, [0, 0, 1, -750, 7906])
            sage: E.has_rational_cm()
            True
            sage: E.cm_discriminant()
            -27
            sage: points = [E([10, -38]), E([15/49*a + 760/49, 675/343*a - 884/343])]
            sage: E.saturation(points, verbose=True) # long time (17s)
            Computing lower height bound..
            ..done: 7.168735020029907e-06
            p-saturating for primes p < 132
            ...
            Saturating at p=131
            --starting full 131-saturation
            Using sieve method to saturate...
            E has 131-torsion over Finite Field of size 617011, projecting points
            --> [(10 : 616973 : 1), (592083 : 192224 : 1)]
            --rank is now 1
            --rank is now 2
            Reached full rank: points were 131-saturated
            Points were 131-saturated
            --already 131-saturated
            ([(10 : -38 : 1), (15/49*a + 760/49 : 675/343*a - 884/343 : 1)],
            1,
            0.123378097374749)
        """
        verbose = self._verbose
        # This code does a lot of elliptic curve group structure
        # computations, and would benefit immensely if those were sped
        # up.
        n = len(Plist)
        if n == 0:
            return False

        if n == 1 and p == 2:
            pts = Plist[0].division_points(p)
            if pts:
                return (0, pts[0])
            else:
                return False

        E = self._curve

        if not sieve:
            from sage.groups.generic import multiples
            from sage.schemes.projective.projective_space import ProjectiveSpace

            mults = [list(multiples(P, p)) for P in Plist[:-1]] + [list(multiples(Plist[-1],2))]
            E0 = E(0)

            for v in ProjectiveSpace(GF(p),n-1): # an iterator
                w = tuple(int(x) for x in v)
                P = sum([m[c] for m,c in zip(mults,w)],E0)
                pts = P.division_points(p)
                if pts:
                    if verbose:
                        print("  points not saturated at {}, increasing index by {}".format(p,p))
                        # w will certainly have a coordinate equal to 1
                    return (w.index(1), pts[0])
            # we only get here if no linear combination is divisible by p,
            # so the points are p-saturated:
            return False

        # Now we use the more sophisticated sieve to either prove
        # p-saturation, or compute a much smaller list of possible points
        # to test for p-divisibility:

        if verbose:
            print("Using sieve method to saturate...")
        from sage.matrix.constructor import matrix
        from sage.sets.primes import Primes

        A = matrix(GF(p), 0, n)
        rankA = 0
        count = 0

        # We reduce the curve modulo primes Q of degree 1 above
        # rational primes q chosen as follows: we avoid primes of bad
        # reduction and primes dividing the discriminant of the
        # defining polynomial of the field, so that we can avoid
        # constructing number field primes entirely and just look for
        # roots amodq of the defining polynomial mod q, then reduction
        # of the curve is easy.

        # We also avoid primes dividing the denominators of any of the
        # points: the points would reduce to 0 modulo such primes
        # anyway, and this way reduction of the points is also easy.

        # Lastly, there is a special case when E has rational CM by
        # some discriminant D.  In the non-split case (kro(D,p)=-1)
        # the image of the mod-p Galois representation is cyclic of
        # order p^2-1 and there will be no p-torsion mod Q unless
        # Frob(Q) has trivial image; a necessary condition for this is
        # q=1 (mod p) so we restrict to such q.  That way the density
        # of q for which there is a point of order p is 1/(p+1)
        # instead of 1/(p^2-1).  This test was put in after running
        # many tests: for example, LMFDB curve 2.0.3.1-50625.1-CMb2
        # has saturation index bound 132 and to saturate at p=131
        # requires q=617011. (In the split case the density is 1/(p-1)
        # and there is no simple test.)

        avoid = [self._N, self._D] + [P[0].denominator_ideal().norm() for P in Plist]
        cm_test = E.has_rational_cm() and kro(E.cm_discriminant(), p) == -1
        for q in Primes():
            if any(q.divides(m) for m in avoid):
                continue
            if cm_test and not p.divides(q-1):
                continue
            self.add_reductions(q) # does nothing if key q is already there
            for amodq in self._reductions[q]:
                (nq, Eq) = self._reductions[q][amodq]
                if not p.divides(nq):
                    continue
                if verbose:
                    print("E has %s-torsion over %s, projecting points" % (p,GF(q)))
                projPlist = [Eq([reduce_mod_q(c, amodq) for c in pt]) for pt in Plist]
                if verbose:
                    print(" --> %s" % projPlist)
                try:
                    vecs = p_projections(Eq, projPlist, p)
                except ValueError:
                    vecs = []
                for v in vecs:
                    A = matrix(A.rows()+[v])
                    newrank = A.rank()
                    if verbose:
                        print(" --rank is now %s" % newrank)
                    if newrank == n:
                        if verbose:
                            print("Reached full rank: points were %s-saturated" % p)
                        return False
                    if newrank == rankA:
                        count += 1
                        if count == 15:
                            if verbose:
                                print("! rank same for 15 steps, checking kernel...")
                            # no increase in rank for the last 15 primes Q
                            # find the points in the kernel and call the no-sieve version
                            vecs = A.right_kernel().basis()
                            if verbose:
                                print("kernel vectors: %s" % vecs)
                            Rlist = [sum([int(vi)*Pi for vi,Pi in zip(v,Plist)],E(0))
                                     for v in vecs]
                            if verbose:
                                print("points generating kernel: %s" % Rlist)

                            # If the nullity of A were 1 (the usual
                            # case) we take any nonzero vector in its
                            # kernel and compute that linear
                            # combination of the points, giving a
                            # point which is certainly a p-multiple
                            # modulo 15 primes Q, and we test if it
                            # actually is a p-multiple:
                            if len(Rlist) == 1:
                                R = Rlist[0]
                                pts = R.division_points(p)
                                if pts:
                                    pt = pts[0]
                                    v = vecs[0]
                                    # replace one of the original
                                    # points with this one; we can
                                    # replace any for which the
                                    # coefficient of v is nonzero
                                    if verbose:
                                        print("-- points were not {}-saturated, gaining index {}".format(p,p))
                                    j = next(i for i,x in enumerate(v) if x)
                                    return (j, pt)
                                else:
                                    # R is not a p-multiple so the
                                    # points were p-saturated
                                    return False

                            # Else we call the non-sieve version with
                            # a list of points which are all
                            # p-multiples modulo 15 primes, and we
                            # will just try to divide all linear
                            # combinations of them
                            res = self.p_saturation(Rlist, p, sieve=False)
                            if res:
                                jj, R = res
                                v = vecs[jj]
                                # The v-linear combination of Rlist
                                # really is p*R.  Now to enlarge the
                                # span, we may replce the j'th point
                                # in Plist with R, where v[j] is
                                # nonzero.
                                if verbose:
                                    print("-- points were not {}-saturated, gaining index {}".format(p,p))
                                j = next(i for i,x in enumerate(v) if x)
                                return (j, R)
                            else:
                                # points really were saturated
                                if verbose:
                                    print("-- points were %s-saturated" % p)
                                return False
                    else: # rank went up but is <n; carry on using more Qs
                        rankA = newrank
                        count = 0


def p_projections(Eq, Plist, p, debug=False):
    r"""

    INPUT:

    - ``Eq`` -- an elliptic curve over a finite field

    - ``Plist`` -- list of points on `Eq`

    - ``p`` -- a prime number

    OUTPUT:

    A list of `r\le2` vectors in `\GF{p^n}`, the images of the points in
    `G \otimes \GF{p}`, where `r` is the number of vectors is the
    `p`-rank of `Eq`.

    ALGORITHM:

    First project onto the `p`-primary part of `Eq`.  If that has
    `p`-rank 1 (i.e. is cyclic), use discrete logs there to define a
    map to `\GF{p}`, otherwise use the Weil pairing to define two
    independent maps to `\GF{p}`.

    EXAMPLES:

    This curve has three independent rational points::

        sage: E = EllipticCurve([0,0,1,-7,6])

    We reduce modulo `409` where its order is `3^2\cdot7^2`; the
    `3`-primary part is non-cyclic while the `7`-primary part is
    cyclic of order `49`::

        sage: F = GF(409)
        sage: EF = E.change_ring(F)
        sage: G = EF.abelian_group()
        sage: G
        Additive abelian group isomorphic to Z/147 + Z/3
         embedded in Abelian group of points on Elliptic Curve
          defined by y^2 + y = x^3 + 402*x + 6 over Finite Field of size 409
        sage: G.order().factor()
        3^2 * 7^2

    We construct three points and project them to the `p`-primary
    parts for `p=2,3,5,7`, yielding 0,2,0,1 vectors of length 3 modulo
    `p` respectively.  The exact vectors output depend on the computed
    generators of `G`::

        sage: Plist = [EF([-2,3]), EF([0,2]), EF([1,0])]
        sage: from sage.schemes.elliptic_curves.saturation import p_projections
        sage: [(p, p_projections(EF, Plist, p)) for p in primes(11)]  # random
        [(2, []), (3, [(0, 2, 2), (2, 2, 1)]), (5, []), (7, [(5, 1, 1)])]
        sage: [(p, len(p_projections(EF, Plist, p))) for p in primes(11)]
        [(2, 0), (3, 2), (5, 0), (7, 1)]
    """
    if debug:
        print("In p_projections(Eq,Plist,p) with Eq = {}, Plist = {}, p = {}".format(Eq,Plist,p))
    n = Eq.cardinality()
    m = n.prime_to_m_part(p)      # prime-to-p part of order
    if debug:
        print("m={}, n={}".format(m,n))
    if m == n: # p-primary part trivial, nothing to do
        return []
    G = Eq.abelian_group()
    if debug:
        print("gens = {}".format(G.gens()))

    # project onto p-primary part

    pts = [m*pt for pt in Plist]
    gens = [m*g.element() for g in G.gens()]
    gens = [g for g in gens if g]
    if debug:
        print("gens for {}-primary part of G: {}".format(p, gens))
        print("{}*points: {}".format(m,pts))
    from sage.groups.generic import discrete_log as dlog
    from sage.modules.free_module_element import vector
    Fp = GF(p)

    # If the p-primary part is cyclic we use elliptic discrete logs directly:

    if len(gens) == 1:
        g = gens[0]
        pp = g.order()
        if debug:
            print("Cyclic case, taking dlogs to base {} of order {}".format(g,pp))
        # logs are well-defined mod pp, hence mod p
        v = [dlog(pt, g, ord=pp, operation='+') for pt in pts]
        if debug:
            print("dlogs: {}".format(v))
        return [vector(Fp, v)]

    # We make no assumption about which generator order divides the
    # other, since conventions differ!

    orders = [g.order() for g in gens]
    p1, p2 = min(orders), max(orders)
    g1, g2 = gens
    if debug:
        print("Non-cyclic case, orders = {}, p1={}, p2={}, g1={}, g2={}".format(orders,p1,p2,g1,g2))

    # Now the p-primary part of the reduction is non-cyclic of
    # exponent p2, and we use the Weil pairing, whose values are p1'th
    # roots of unity with p1|p2, together with discrete log in the
    # multiplicative group.

    zeta = g1.weil_pairing(g2, p2)  # a primitive p1'th root of unity
    if debug:
        print("wp of gens = {} with order {}".format(zeta, zeta.multiplicative_order()))
        assert zeta.multiplicative_order() == p1, "Weil pairing error during saturation: p={}, G={}, Plist={}".format(p, G, Plist)

    # logs are well-defined mod p1, hence mod p

    return [vector(Fp, [dlog(pt.weil_pairing(g1, p2), zeta,
                             ord=p1, operation='*') for pt in pts]),
            vector(Fp, [dlog(pt.weil_pairing(g2, p2), zeta,
                             ord=p1, operation='*') for pt in pts])]



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/dirichlet.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    int dirichlet_group_init(dirichlet_group_t G, ulong q) noexcept
    void dirichlet_subgroup_init(dirichlet_group_t H, const dirichlet_group_t G, ulong h) noexcept
    void dirichlet_group_clear(dirichlet_group_t G) noexcept
    ulong dirichlet_group_size(const dirichlet_group_t G) noexcept
    ulong dirichlet_group_num_primitive(const dirichlet_group_t G) noexcept
    void dirichlet_group_dlog_precompute(dirichlet_group_t G, ulong num) noexcept
    void dirichlet_group_dlog_clear(dirichlet_group_t G) noexcept
    void dirichlet_char_init(dirichlet_char_t chi, const dirichlet_group_t G) noexcept
    void dirichlet_char_clear(dirichlet_char_t chi) noexcept
    void dirichlet_char_print(const dirichlet_group_t G, const dirichlet_char_t chi) noexcept
    void dirichlet_char_log(dirichlet_char_t x, const dirichlet_group_t G, ulong m) noexcept
    ulong dirichlet_char_exp(const dirichlet_group_t G, const dirichlet_char_t x) noexcept
    ulong _dirichlet_char_exp(dirichlet_char_t x, const dirichlet_group_t G) noexcept
    void dirichlet_char_one(dirichlet_char_t x, const dirichlet_group_t G) noexcept
    void dirichlet_char_first_primitive(dirichlet_char_t x, const dirichlet_group_t G) noexcept
    void dirichlet_char_set(dirichlet_char_t x, const dirichlet_group_t G, const dirichlet_char_t y) noexcept
    int dirichlet_char_next(dirichlet_char_t x, const dirichlet_group_t G) noexcept
    int dirichlet_char_next_primitive(dirichlet_char_t x, const dirichlet_group_t G) noexcept
    ulong dirichlet_index_char(const dirichlet_group_t G, const dirichlet_char_t x) noexcept
    void dirichlet_char_index(dirichlet_char_t x, const dirichlet_group_t G, ulong j) noexcept
    bint dirichlet_char_eq(const dirichlet_char_t x, const dirichlet_char_t y) noexcept
    int dirichlet_char_eq_deep(const dirichlet_group_t G, const dirichlet_char_t x, const dirichlet_char_t y) noexcept
    bint dirichlet_char_is_principal(const dirichlet_group_t G, const dirichlet_char_t chi) noexcept
    ulong dirichlet_conductor_ui(const dirichlet_group_t G, ulong a) noexcept
    ulong dirichlet_conductor_char(const dirichlet_group_t G, const dirichlet_char_t x) noexcept
    int dirichlet_parity_ui(const dirichlet_group_t G, ulong a) noexcept
    int dirichlet_parity_char(const dirichlet_group_t G, const dirichlet_char_t x) noexcept
    ulong dirichlet_order_ui(const dirichlet_group_t G, ulong a) noexcept
    ulong dirichlet_order_char(const dirichlet_group_t G, const dirichlet_char_t x) noexcept
    bint dirichlet_char_is_real(const dirichlet_group_t G, const dirichlet_char_t chi) noexcept
    bint dirichlet_char_is_primitive(const dirichlet_group_t G, const dirichlet_char_t chi) noexcept
    ulong dirichlet_pairing(const dirichlet_group_t G, ulong m, ulong n) noexcept
    ulong dirichlet_pairing_char(const dirichlet_group_t G, const dirichlet_char_t chi, const dirichlet_char_t psi) noexcept
    ulong dirichlet_chi(const dirichlet_group_t G, const dirichlet_char_t chi, ulong n) noexcept
    void dirichlet_chi_vec(ulong * v, const dirichlet_group_t G, const dirichlet_char_t chi, slong nv) noexcept
    void dirichlet_chi_vec_order(ulong * v, const dirichlet_group_t G, const dirichlet_char_t chi, ulong order, slong nv) noexcept
    void dirichlet_char_mul(dirichlet_char_t chi12, const dirichlet_group_t G, const dirichlet_char_t chi1, const dirichlet_char_t chi2) noexcept
    void dirichlet_char_pow(dirichlet_char_t c, const dirichlet_group_t G, const dirichlet_char_t a, ulong n) noexcept
    void dirichlet_char_lift(dirichlet_char_t chi_G, const dirichlet_group_t G, const dirichlet_char_t chi_H, const dirichlet_group_t H) noexcept
    void dirichlet_char_lower(dirichlet_char_t chi_H, const dirichlet_group_t H, const dirichlet_char_t chi_G, const dirichlet_group_t G) noexcept



"""
All imports for the Sage (IPython) command line

This is all.py (load all sage functions) plus set-up for the Sage commandline.
"""
# ****************************************************************************
#       Copyright (C) 2007 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
sage_mode = 'cmdline'

from sage.all import *

try:
    from sage.calculus.predefined import x
except ImportError:
    pass

from sage.misc.lazy_import import lazy_import

for pkg in ['axiom', 'fricas', 'gap', 'gap3', 'giac', 'gp',
            'gnuplot', 'kash', 'magma', 'macaulay2', 'maple',
            'mathematica', 'mathics3', 'matlab',
            'mupad', 'mwrank', 'octave', 'qepcad', 'singular',
            'sage0', 'lie', 'r']:
    lazy_import(f'sage.interfaces.{pkg}', f'{pkg}_console')
del pkg

lazy_import('sage.interfaces.maxima_abstract', 'maxima_console')

sage.misc.session.init()



# sage_setup: distribution = sagemath-modules
r"""
Index of bounds on the parameters of codes

The :obj:`codes.bounds` object may be used to access the bounds that Sage can compute.

{INDEX_OF_FUNCTIONS}

To import these names into the global namespace, use::

    sage: from sage.coding.bounds_catalog import *
"""
from sage.misc.lazy_import import lazy_import as _lazy_import
_lazy_import("sage.coding.code_bounds", ["codesize_upper_bound",
            "dimension_upper_bound",
            "volume_hamming",
            "gilbert_lower_bound",
            "plotkin_upper_bound",
            "griesmer_upper_bound",
            "elias_upper_bound",
            "hamming_upper_bound",
            "singleton_upper_bound",
            "gv_info_rate",
            "entropy",
            "gv_bound_asymp",
            "hamming_bound_asymp",
            "singleton_bound_asymp",
            "plotkin_bound_asymp",
            "elias_bound_asymp",
            "mrrw1_bound_asymp"])

_lazy_import("sage.coding.delsarte_bounds",
            ["krawtchouk",
             "eberlein",
             "delsarte_bound_constant_weight_code",
             "delsarte_bound_hamming_space",
             "delsarte_bound_additive_hamming_space",
             "delsarte_bound_Q_matrix"])


from sage.misc.rest_index_of_methods import gen_rest_table_index as _gen_rest_table_index
import sys as _sys
__doc__ = __doc__.format(INDEX_OF_FUNCTIONS=_gen_rest_table_index(_sys.modules[__name__], only_local_functions=False))



# sage_setup: distribution = sagemath-modules
r"""
Base class for dense matrices

TESTS::

    sage: R.<a,b> = QQ[]
    sage: m = matrix(R,2,[0,a,b,b^2])
    sage: TestSuite(m).run(skip='_test_minpoly')
"""

cimport sage.matrix.matrix as matrix

from sage.structure.richcmp cimport richcmp_item, rich_to_bool
import sage.matrix.matrix_space
import sage.structure.sequence


cdef class Matrix_dense(matrix.Matrix):
    cdef bint is_sparse_c(self) noexcept:
        return 0

    cdef bint is_dense_c(self) noexcept:
        return 1

    def __copy__(self):
        """
        Return a copy of this matrix. Changing the entries of the copy will
        not change the entries of this matrix.
        """
        A = self.new_matrix(entries=self.list(), coerce=False, copy=False)
        if self._subdivisions is not None:
            A.subdivide(*self.subdivisions())
        return A

    cdef void set_unsafe_int(self, Py_ssize_t i, Py_ssize_t j, int value) noexcept:
        self.set_unsafe(i, j, value)

    def _pickle(self):
        version = -1
        data = self._list()  # linear list of all elements
        return data, version

    def _unpickle_generic(self, data, int version):
        cdef Py_ssize_t i, j, k
        if version == -1:
            # data is a *list* of the entries of the matrix.
            # TODO: Change the data[k] below to use the fast list access macros from the Python/C API
            k = 0
            for i from 0 <= i < self._nrows:
                for j from 0 <= j < self._ncols:
                    self.set_unsafe(i, j, data[k])
                    k = k + 1
        else:
            raise RuntimeError("unknown matrix version (=%s)" % version)

    cpdef _richcmp_(self, right, int op):
        """
        EXAMPLES::

            sage: P.<x> = QQ[]
            sage: m = matrix([[x,x+1],[1,x]])
            sage: n = matrix([[x+1,x],[1,x]])
            sage: o = matrix([[x,x],[1,x]])
            sage: m < n
            True
            sage: m == m
            True
            sage: n > m
            True
            sage: m <= o
            False

        TESTS:

        Check :issue:`27629`::

            sage: # needs sage.symbolic
            sage: var('x')
            x
            sage: assume(x, 'real')
            sage: M = matrix([[0, -x], [x, 0]])
            sage: M.transpose() == M
            False
        """
        other = <Matrix_dense>right
        cdef Py_ssize_t i, j
        # Parents are equal, so dimensions of self and other are equal
        for i in range(self._nrows):
            for j in range(self._ncols):
                lij = self.get_unsafe(i, j)
                rij = other.get_unsafe(i, j)
                r = richcmp_item(lij, rij, op)
                if r is not NotImplemented:
                    return bool(r)
        # Matrices are equal
        return rich_to_bool(op, 0)

    def transpose(self):
        """
        Return the transpose of ``self``, without changing ``self``.

        EXAMPLES: We create a matrix, compute its transpose, and note that
        the original matrix is not changed.

        ::

            sage: M = MatrixSpace(QQ,  2)
            sage: A = M([1,2,3,4])
            sage: B = A.transpose()
            sage: print(B)
            [1 3]
            [2 4]
            sage: print(A)
            [1 2]
            [3 4]

        :attr:`~sage.matrix.matrix2.Matrix.T` is a convenient shortcut for the transpose::

           sage: A.T
           [1 3]
           [2 4]

        ::

            sage: A.subdivide(None, 1); A
            [1|2]
            [3|4]
            sage: A.transpose()
            [1 3]
            [---]
            [2 4]
        """
        (nc, nr) = (self.ncols(), self.nrows())
        cdef Matrix_dense trans
        trans = self.new_matrix(nrows = nc, ncols = nr,
                                copy=False, coerce=False)

        cdef Py_ssize_t i, j

        if type(trans) == type(self):
            for j from 0<= j < nc:
                for i from 0<= i < nr:
                    trans.copy_from_unsafe(j, i, self, i, j)
        else:
            for j from 0<= j < nc:
                for i from 0<= i < nr:
                    trans.set_unsafe(j,i,self.get_unsafe(i,j))

        if self._subdivisions is not None:
            row_divs, col_divs = self.subdivisions()
            trans.subdivide(col_divs, row_divs)
        return trans

    def antitranspose(self):
        """
        Return the antitranspose of ``self``, without changing ``self``.

        EXAMPLES::

            sage: A = matrix(2,3,range(6)); A
            [0 1 2]
            [3 4 5]
            sage: A.antitranspose()
            [5 2]
            [4 1]
            [3 0]

        ::

            sage: A.subdivide(1,2); A
            [0 1|2]
            [---+-]
            [3 4|5]
            sage: A.antitranspose()
            [5|2]
            [-+-]
            [4|1]
            [3|0]
        """
        nc, nr = self.ncols(), self.nrows()
        cdef Matrix_dense atrans
        atrans = self.new_matrix(nrows=nc, ncols=nr,
                                 copy=False, coerce=False)
        cdef Py_ssize_t i, j
        cdef Py_ssize_t ri, rj # reversed i and j
        rj = nc

        if type(atrans) == type(self):
            for j from 0 <= j < nc:
                ri = nr
                rj -= 1
                for i from 0 <= i < nr:
                    ri -= 1
                    atrans.copy_from_unsafe(j, i, self, ri, rj)
        else:
            for j from 0 <= j < nc:
                ri = nr
                rj -= 1
                for i from 0 <= i < nr:
                    ri -= 1
                    atrans.set_unsafe(j, i, self.get_unsafe(ri, rj))

        if self._subdivisions is not None:
            row_divs, col_divs = self.subdivisions()
            atrans.subdivide([nc - t for t in reversed(col_divs)],
                             [nr - t for t in reversed(row_divs)])
        return atrans

    def _reverse_unsafe(self):
        r"""
        TESTS::

            sage: m = matrix(QQ, 2, 3, range(6))
            sage: m._reverse_unsafe()
            sage: m
            [5 4 3]
            [2 1 0]
        """
        cdef Py_ssize_t i, j
        cdef Py_ssize_t nrows = self._nrows
        cdef Py_ssize_t ncols = self._ncols
        for i in range(nrows // 2):
            for j in range(ncols):
                e1 = self.get_unsafe(i, j)
                e2 = self.get_unsafe(nrows - i - 1, ncols - j - 1)
                self.set_unsafe(i, j, e2)
                self.set_unsafe(nrows - i - 1, ncols - j - 1, e1)
        if nrows % 2:
            i = nrows // 2
            for j in range(ncols // 2):
                e1 = self.get_unsafe(i, j)
                e2 = self.get_unsafe(nrows - i - 1, ncols - j - 1)
                self.set_unsafe(i, j, e2)
                self.set_unsafe(nrows - i - 1, ncols - j - 1, e1)

    def _elementwise_product(self, right):
        r"""
        Return the elementwise product of two dense
        matrices with identical base rings.

        This routine assumes that ``self`` and ``right``
        are both matrices, both dense, with identical
        sizes and with identical base rings.  It is
        "unsafe" in the sense that these conditions
        are not checked and no sensible errors are
        raised.

        This routine is meant to be called from the
        :meth:`~sage.matrix.matrix2.Matrix.elementwise_product`
        method, which will ensure that this routine receives
        proper input.  More thorough documentation is provided
        there.

        EXAMPLES::

            sage: A = matrix(ZZ, 2, range(6), sparse=False)
            sage: B = matrix(ZZ, 2, [1,0,2,0,3,0], sparse=False)
            sage: A._elementwise_product(B)
            [ 0  0  4]
            [ 0 12  0]

        AUTHOR:

        - Rob Beezer (2009-07-14)
        """
        cdef Py_ssize_t r, c
        cdef Matrix_dense other, prod

        nc, nr = self.ncols(), self.nrows()
        other = right
        prod = self.new_matrix(nr, nc, copy=False, coerce=False)
        for r in range(nr):
            for c in range(nc):
                entry = self.get_unsafe(r, c)*other.get_unsafe(r, c)
                prod.set_unsafe(r, c, entry)
        return prod

    def _derivative(self, var=None, R=None):
        """
        Differentiate with respect to var by differentiating each element
        with respect to var.

        .. SEEALSO::

           :meth:`derivative`

        EXAMPLES::

            sage: m = matrix(2, [x^i for i in range(4)])                                # needs sage.symbolic
            sage: m._derivative(x)                                                      # needs sage.symbolic
            [    0     1]
            [  2*x 3*x^2]

        TESTS:

        Verify that :issue:`15067` is fixed::

            sage: u = matrix(1, 2, [-1, 1])
            sage: derivative(u, x)                                                      # needs sage.symbolic
            [0 0]
        """
        # We could just use apply_map
        if self._nrows==0 or self._ncols==0:
            return self.__copy__()
        v = [sage.calculus.functional.derivative(z, var) for z in self.list()]
        if R is None:
            v = sage.structure.sequence.Sequence(v)
            R = v.universe()
        M = sage.matrix.matrix_space.MatrixSpace(R, self._nrows,
                   self._ncols, sparse=False)
        image = M(v)
        if self._subdivisions is not None:
            image.subdivide(*self.subdivisions())
        return image

    def _multiply_classical(left, matrix.Matrix right):
        """
        Multiply the matrices left and right using the classical `O(n^3)`
        algorithm.

        This method will almost always be overridden either by the
        implementation in :class:`~sage.matrix.Matrix_generic_dense`) or by
        more specialized versions, but having it here makes it possible to
        implement specialized dense matrix types with their own data structure
        without necessarily implementing ``_multiply_classical``, as described
        in :mod:`sage.matrix.docs`.

        TESTS::

            sage: from sage.matrix.matrix_dense import Matrix_dense
            sage: mats = [
            ....:     matrix(2, 2, [1, 2, 3, 4]),
            ....:     matrix(2, 1, [1, 2]),
            ....:     matrix(3, 2, [1, 2, 3, 4, 5, 6]),
            ....:     matrix(ZZ, 0, 2),
            ....:     matrix(ZZ, 2, 0)
            ....: ]
            sage: all(Matrix_dense._multiply_classical(a, b) == a*b
            ....:     for a in mats for b in mats if a.ncols() == b.nrows())
            True
            sage: Matrix_dense._multiply_classical(matrix(2, 1), matrix(2, 0))
            Traceback (most recent call last):
            ...
            ArithmeticError: number of columns of left must equal number of rows of right
        """
        cdef Py_ssize_t i, j, k
        if left._ncols != right._nrows:
            raise ArithmeticError("number of columns of left must equal number of rows of right")
        zero = left.base_ring().zero()
        cdef matrix.Matrix res = left.new_matrix(nrows=left._nrows, ncols=right._ncols)
        for i in range(left._nrows):
            for j in range(right._ncols):
                dotp = zero
                for k in range(left._ncols):
                    dotp += left.get_unsafe(i, k) * right.get_unsafe(k, j)
                res.set_unsafe(i, j, dotp)
        return res



# sage_setup: distribution = sagemath-pari
# sage.doctest: needs sage.rings.padics
r"""
Tools for creation of `p`-adic elements

The functions in this file are used in creating new `p`-adic elements.

When creating a `p`-adic element, the user can specify that the absolute
precision be bounded and/or that the relative precision be bounded.
Moreover, different `p`-adic parents impose their own bounds on the
relative or absolute precision of their elements.  The precision
determines to what power of `p` the defining data will be reduced, but
the valuation of the resulting element needs to be determined before
the element is created.  Moreover, some defining data can impose their
own precision bounds on the result.

AUTHORS:

- David Roe (2012-03-01)
"""

# ****************************************************************************
#       Copyright (C) 2007-2013 David Roe <roed.math@gmail.com>
#                               William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cpython.long cimport *
from sage.ext.stdsage cimport PY_NEW
from sage.libs.gmp.all cimport *
from sage.arith.rational_reconstruction cimport mpq_rational_reconstruction
from sage.rings.integer cimport Integer
from sage.rings.rational cimport Rational
from sage.rings.padics.padic_generic_element cimport pAdicGenericElement
import sage.rings.finite_rings.integer_mod
from cypari2.types cimport *
from cypari2.gen cimport Gen as pari_gen
from sage.rings.infinity import infinity
from sage.structure.element cimport parent


cdef long maxordp = (1L << (sizeof(long) * 8 - 2)) - 1
cdef long minusmaxordp = -maxordp
# The following Integer (resp. Rational) is used so that
# the functions here don't need to initialize an mpz_t (resp. mpq_t)
cdef Integer temp = PY_NEW(Integer)
cdef Rational rat_temp = PY_NEW(Rational)

cdef long get_ordp(x, PowComputer_class prime_pow) except? -10000:
    """
    Determine the valuation of the `p`-adic element
    that will result from the given data ``x``.

    Note that the valuation can differ depending on the ring: if the
    new `p`-adic element is being created in a ring with ramification
    then the valuation will be larger.

    Some preprocessing is done in the initialization methods of the element, so
    the type of x is restricted.

    Also note that for some kinds of elements conversion to and from
    Integers and Rationals is done in a custom morphism rather than
    through this function.

    INPUT:

    - ``x`` -- data defining a new `p`-adic element: a Python int, an
      Integer, Rational, an element of Zp or Qp with the same prime, a
      PARI `p`-adic element, a list, a tuple, or an IntegerMod.

    - a PowComputer associated to a `p`-adic ring, which determines
      the prime and the ramification degree.

    OUTPUT:

    - a long, giving the valuation of the resulting `p`-adic element.
      If the input is zero, returns ``maxordp``
    """
    cdef long k, n, p, curterm, shift, f, ratio, e = prime_pow.e
    cdef Integer value
    cdef GEN pari_tmp
    if isinstance(x, int):
        if x == 0:
            return maxordp
        try:
            n = PyLong_AsLong(x)
        except OverflowError:
            return get_ordp(Integer(x), prime_pow)
        else:
            if mpz_fits_slong_p(prime_pow.prime.value) == 0:
                # x is not divisible by p
                return 0
            p = mpz_get_si(prime_pow.prime.value)
            k = 0
            while n % p == 0:
                k += 1
                n = n / p
    elif isinstance(x, Integer):
        if mpz_sgn((<Integer>x).value) == 0:
            return maxordp
        k = mpz_remove(temp.value, (<Integer>x).value, prime_pow.prime.value)
    elif isinstance(x, Rational):
        if mpq_sgn((<Rational>x).value) == 0:
            return maxordp
        k = mpz_remove(temp.value, mpq_numref((<Rational>x).value), prime_pow.prime.value)
        if k == 0:
            k = -mpz_remove(temp.value, mpq_denref((<Rational>x).value), prime_pow.prime.value)
    elif isinstance(x, (list, tuple)):
        f = prime_pow.f
        if (e == 1 and len(x) > f) or (e != 1 and len(x) > e):
            # could reduce modulo the defining polynomial but that isn't currently supported
            raise ValueError("List too long")
        k = maxordp
        shift = 0
        for a in x:
            if isinstance(a, (list, tuple)):
                if e == 1 or f == 1:
                    raise ValueError("nested lists not allowed for unramified and eisenstein extensions")
                for b in a:
                    if isinstance(b, (list, tuple)):
                        raise ValueError("list nesting too deep")
                    curterm = get_ordp(b, prime_pow)
                    k = min(k, curterm + shift, maxordp)
            else:
                curterm = get_ordp(a, prime_pow)
                k = min(k, curterm + shift, maxordp)
            if e != 1:
                shift += 1
        # We don't want to multiply by e again.
        return k
    elif isinstance(x, pAdicGenericElement):
        if x.parent().is_relaxed():
            return x.valuation()
        k = (<pAdicGenericElement>x).valuation_c()
        if not (<pAdicGenericElement>x)._is_base_elt(prime_pow.prime):
            # We have to be careful with overflow
            ratio = e // x.parent().absolute_e()
            if k >= maxordp // ratio:
                return maxordp
            elif k <= minusmaxordp // ratio:
                return minusmaxordp
            else:
                return (k*e) // x.parent().absolute_e()
        return k
    elif isinstance(x, pari_gen):
        pari_tmp = (<pari_gen>x).g
        if typ(pari_tmp) == t_PADIC:
            k = valp(pari_tmp)
        else: # t_INT and t_FRAC were converted before this function
            raise TypeError("unsupported coercion from pari: only p-adics, integers and rationals allowed")
    elif isinstance(x, sage.rings.finite_rings.integer_mod.IntegerMod_abstract):
        value = <Integer>x.lift()
        if mpz_sgn(value.value) == 0:
            return maxordp
        k = mpz_remove(temp.value, value.value, prime_pow.prime.value)
    else:
        raise NotImplementedError("Cannot determine p-adic valuation of an element of %s" % parent(x))
    # Should check for overflow
    return k * e

cdef long get_preccap(x, PowComputer_class prime_pow) except? -10000:
    """
    Determine the maximum absolute precision possible
    for an element created from the given data ``x``.

    Note that the valuation can differ depending on the ring: if the
    new `p`-adic element is being created in a ring with ramification
    then the precision bound will be larger.

    Some preprocessing is done in the initialization methods of the element, so
    the type of x is restricted.

    Also note that for some kinds of elements conversion to and from
    Integers and Rationals is done in a custom morphism rather than
    through this function.

    INPUT:

    - ``x`` -- data defining a new `p`-adic element: an Integer,
      Rational, an element of Zp or Qp with the same prime, a PARI
      `p`-adic element, a list, a tuple, or an IntegerMod.
    - ``prime_pow`` -- the PowComputer for the ring into which ``x``
      is being converted.  This is used to determine the prime and the
      ramification degree.

    OUTPUT:

    - a long, giving the absolute precision modulo which the input is
      defined.  If the input is exact, returns ``maxordp``
    """
    cdef long k, shift, e = prime_pow.e
    cdef Integer prec
    cdef GEN pari_tmp
    if isinstance(x, (int, Integer, Rational)):
        return maxordp
    elif isinstance(x, (list, tuple)):
        k = maxordp
        shift = 0
        for a in x:
            if isinstance(a, (list, tuple)):
                for b in a:
                    curterm = get_preccap(b, prime_pow)
                    k = min(k, curterm + shift)
            else:
                curterm = get_preccap(a, prime_pow)
                k = min(k, curterm + shift)
            if e != 1:
                shift += 1
        # We don't want to multiply by e again.
        return k
    elif isinstance(x, pAdicGenericElement):
        if (<pAdicGenericElement>x)._is_exact_zero():
            return maxordp
        prec = <Integer>x.precision_absolute()
        if prec is infinity:
            return maxordp
        k = mpz_get_si(prec.value)
        if not (<pAdicGenericElement>x)._is_base_elt(prime_pow.prime):
            # since x lives in a subfield, the ramification index of x's parent will divide e.
            return (k * e) // x.parent().absolute_e()
    elif isinstance(x, pari_gen):
        pari_tmp = (<pari_gen>x).g
        # since get_ordp has been called typ(x.g) == t_PADIC
        k = valp(pari_tmp) + precp(pari_tmp)
    elif isinstance(x, sage.rings.finite_rings.integer_mod.IntegerMod_abstract):
        k = mpz_remove(temp.value, (<Integer>x.modulus()).value, prime_pow.prime.value)
        if mpz_cmp_ui(temp.value, 1) != 0:
            raise TypeError("cannot coerce from the given integer mod ring (not a power of the same prime)")
    else:
        raise NotImplementedError("Cannot determine p-adic precision of an element of %s" % parent(x))
    return k * e

cdef long comb_prec(iprec, long prec) except? -10000:
    """
    Return the minimum of ``iprec`` and ``prec``.

    INPUT:

    - ``iprec`` -- infinity, integer or something that can be converted to an
      integer

    - ``prec`` -- a long
    """
    if iprec is infinity:
        return prec
    cdef Integer intprec
    if isinstance(iprec, Integer):
        intprec = <Integer>iprec
        if mpz_cmp_si(intprec.value, prec) >= 0:
            return prec
        if mpz_fits_slong_p(intprec.value) == 0:
            raise OverflowError("precision overflow")
        return mpz_get_si(intprec.value)
    if isinstance(iprec, int):
        return min(PyLong_AsLong(iprec), prec)
    return comb_prec(Integer(iprec), prec)

cdef int _process_args_and_kwds(long *aprec, long *rprec, args, kwds, bint absolute, PowComputer_class prime_pow) except -1:
    """
    Obtain values for ``absprec`` and ``relprec`` from a
    combination of positional and keyword arguments.

    When creating a `p`-adic element, the user can pass in two arguments: ``absprec`` and ``relprec``.
    In implementing morphisms to speed up conversion from Integers and Rationals,
    we need to determine ``absprec`` and ``relprec`` from the ``args`` and ``kwds`` arguments of
    ``_call_with_args``.  This function collects the code to do so.

    INPUT:

    - ``args`` -- tuple of positional arguments (at most two)

    - ``kwds`` -- dictionary of keyword arguments (only
      ``'relprec'`` and ``'absprec'`` are used)

    - ``absolute`` -- boolean; ``True`` if the precision cap of the ring
      is a cap on absolute precision, ``False`` if a cap on relative
      precision

    - ``prime_pow`` -- a
      :class:`sage.rings.padics.pow_computer.PowComputer_class`
      instance

    OUTPUT:

    - ``aprec`` -- (first argument) the maximum absolute precision of
      the resulting element

    - ``rprec`` -- (second argument) the maximum relative precision of
      the resulting element

    - error status
    """
    if "empty" in kwds:
        # For backward compatibility
        aprec[0] = 0
        rprec[0] = 0
        return 0
    if len(args) > 2:
        raise TypeError("too many positional arguments")
    if len(args) == 2:
        if "relprec" in kwds:
            raise TypeError("_call_with_args() got multiple values for keyword argument 'relprec'")
        relprec = args[1]
    else:
        relprec = kwds.get("relprec", infinity)
    if len(args) >= 1:
        if "absprec" in kwds:
            raise TypeError("_call_with_args() got multiple values for keyword argument 'absprec'")
        absprec = args[0]
    else:
        absprec = kwds.get("absprec", infinity)
    if absolute:
        aprec[0] = comb_prec(absprec, prime_pow.ram_prec_cap)
        rprec[0] = comb_prec(relprec, maxordp)
    else:
        rprec[0] = comb_prec(relprec, prime_pow.ram_prec_cap)
        aprec[0] = comb_prec(absprec, maxordp)

cdef inline long cconv_mpq_t_shared(mpz_t out, mpq_t x, long prec, bint absolute, PowComputer_class prime_pow) except? -10000:
    """
    A fast pathway for conversion of rationals that doesn't require
    precomputation of the valuation.

    INPUT:

    - ``out`` -- an ``mpz_t`` to store the output
    - ``x`` -- an ``mpq_t`` giving the integer to be converted
    - ``prec`` -- a long, giving the precision desired; absolute or
      relative depending on the ``absolute`` input
    - ``absolute`` -- if ``False`` then extracts the valuation and returns
      it, storing the unit in ``out``; if ``True`` then just reduces ``x``
      modulo the precision
    - ``prime_pow`` -- a PowComputer for the ring

    OUTPUT:

    - If ``absolute`` is ``False`` then returns the valuation that was
      extracted (``maxordp`` when `x = 0`).
    """
    cdef long numval, denval
    cdef bint success
    if prec <= 0:
        raise ValueError
    if absolute:
        success = mpz_invert(out, mpq_denref(x), prime_pow.pow_mpz_t_tmp(prec))
        if not success:
            raise ValueError("p divides denominator")
        mpz_mul(out, out, mpq_numref(x))
        mpz_mod(out, out, prime_pow.pow_mpz_t_tmp(prec))
    elif mpq_sgn(x) == 0:
        mpz_set_ui(out, 0)
        return maxordp
    else:
        denval = mpz_remove(out, mpq_denref(x), prime_pow.prime.value)
        mpz_invert(out, out, prime_pow.pow_mpz_t_tmp(prec))
        if denval == 0:
            numval = mpz_remove(temp.value, mpq_numref(x), prime_pow.prime.value)
            mpz_mul(out, out, temp.value)
        else:
            numval = 0
            mpz_mul(out, out, mpq_numref(x))
        mpz_mod(out, out, prime_pow.pow_mpz_t_tmp(prec))
        return numval - denval

cdef inline int cconv_mpq_t_out_shared(mpq_t out, mpz_t x, long valshift, long prec, PowComputer_class prime_pow) except -1:
    """
    Convert the underlying `p`-adic element into a rational.

    INPUT:

    - ``out`` -- gives a rational approximating the input. Currently uses
      rational reconstruction but may change in the future to use a more naive
      method
    - ``x`` -- an ``mpz_t`` giving the underlying `p`-adic element
    - ``valshift`` -- a long giving the power of `p` to shift `x` by
    -` ``prec`` -- a long, the precision of ``x``, used in rational reconstruction
    - ``prime_pow`` -- a PowComputer for the ring
    """
    try:
        mpq_rational_reconstruction(out, x, prime_pow.pow_mpz_t_tmp(prec))
    except (ArithmeticError, ValueError):
        mpz_set(mpq_numref(out), x)
        mpz_set_ui(mpq_denref(out), 1)

    # if valshift is nonzero then we start with x as a p-adic unit,
    # so there will be no powers of p in the numerator or denominator
    # and the following operations yield reduced rationals.
    if valshift > 0:
        mpz_mul(mpq_numref(out), mpq_numref(out), prime_pow.pow_mpz_t_tmp(valshift))
    elif valshift < 0:
        mpz_mul(mpq_denref(out), mpq_denref(out), prime_pow.pow_mpz_t_tmp(-valshift))


cdef inline int cconv_shared(mpz_t out, x, long prec, long valshift, PowComputer_class prime_pow) except -2:
    """
    Conversion from other Sage types.

    INPUT:

    - ``out`` -- an ``mpz_t`` to store the output

    - ``x`` -- a Sage element that can be converted to a `p`-adic element

    - ``prec`` -- a long, giving the precision desired; absolute if
      `valshift = 0`, relative if `valshift != 0`

    - ``valshift`` -- the power of the uniformizer to divide by before
      storing the result in ``out``

    - ``prime_pow`` -- a PowComputer for the ring
    """
    if PyLong_Check(x):
        x = Integer(x)
    elif isinstance(x, pari_gen):
        x = x.sage()
    if isinstance(x, pAdicGenericElement) and x.parent().is_relaxed():
        x = x.lift(valshift + prec)
    elif isinstance(x, pAdicGenericElement) or isinstance(x, sage.rings.finite_rings.integer_mod.IntegerMod_abstract):
        x = x.lift()
    if isinstance(x, Integer):
        if valshift > 0:
            mpz_divexact(out, (<Integer>x).value, prime_pow.pow_mpz_t_tmp(valshift))
            mpz_mod(out, out, prime_pow.pow_mpz_t_tmp(prec))
        elif valshift < 0:
            mpz_mul(out, (<Integer>x).value, prime_pow.pow_mpz_t_tmp(-valshift))
            mpz_mod(out, out, prime_pow.pow_mpz_t_tmp(prec))
        else:
            mpz_mod(out, (<Integer>x).value, prime_pow.pow_mpz_t_tmp(prec))
    elif isinstance(x, Rational):
        if valshift == 0:
            mpz_invert(out, mpq_denref((<Rational>x).value), prime_pow.pow_mpz_t_tmp(prec))
            mpz_mul(out, out, mpq_numref((<Rational>x).value))
        elif valshift < 0:
            mpq_set_z(rat_temp.value, prime_pow.pow_mpz_t_tmp(-valshift))
            mpq_mul(rat_temp.value, rat_temp.value, (<Rational>x).value)
            mpz_invert(out, mpq_denref(rat_temp.value), prime_pow.pow_mpz_t_tmp(prec))
            mpz_mul(out, out, mpq_numref(rat_temp.value))
        else:
            mpz_invert(out, mpq_denref((<Rational>x).value), prime_pow.pow_mpz_t_tmp(prec))
            mpz_divexact(temp.value, mpq_numref((<Rational>x).value), prime_pow.pow_mpz_t_tmp(valshift))
            mpz_mul(out, out, temp.value)
        mpz_mod(out, out, prime_pow.pow_mpz_t_tmp(prec))
    elif isinstance(x, list):
        if valshift == 0:
            if len(x) == 0:
                cconv_shared(out, Integer(0), prec, valshift, prime_pow)
            elif len(x) == 1:
                cconv_shared(out, x[0], prec, valshift, prime_pow)
            else:
                raise NotImplementedError("conversion not implemented from non-prime residue field")
        else:
            raise NotImplementedError
    else:
        raise NotImplementedError("No conversion defined for %s which is a %s in %s" % (x, type(x), x.parent() if hasattr(x, "parent") else "no parent"))

cdef inline long cconv_mpz_t_shared(mpz_t out, mpz_t x, long prec, bint absolute, PowComputer_class prime_pow) except -2:
    """
    A fast pathway for conversion of integers that does not require
    precomputation of the valuation.

    INPUT:

    - ``out`` -- an ``mpz_t`` to store the output
    - ``x`` -- an ``mpz_t`` giving the integer to be converted
    - ``prec`` -- a long, giving the precision desired; absolute or
      relative depending on the ``absolute`` input
    - ``absolute`` -- if ``False`` then extracts the valuation and returns
      it, storing the unit in ``out``; if ``True`` then just reduces ``x``
      modulo the precision
    - ``prime_pow`` -- a PowComputer for the ring

    OUTPUT:

    - If ``absolute`` is False then returns the valuation that was
      extracted (``maxordp`` when `x = 0`).
    """
    cdef long val
    if absolute:
        mpz_mod(out, x, prime_pow.pow_mpz_t_tmp(prec))
    elif mpz_sgn(x) == 0:
        mpz_set_ui(out, 0)
        return maxordp
    else:
        val = mpz_remove(out, x, prime_pow.prime.value)
        mpz_mod(out, out, prime_pow.pow_mpz_t_tmp(prec))
        return val

cdef inline int cconv_mpz_t_out_shared(mpz_t out, mpz_t x, long valshift, long prec, PowComputer_class prime_pow) except -1:
    """
    Convert the underlying `p`-adic element into an integer if possible.

    INPUT:

    - ``out`` -- stores the resulting integer as an integer between 0
      and `p^{prec + valshift}`
    - ``x`` -- an ``mpz_t`` giving the underlying `p`-adic element
    - ``valshift`` -- a long giving the power of `p` to shift `x` by
    -` ``prec`` -- a long, the precision of ``x``; currently not used
    - ``prime_pow`` -- a PowComputer for the ring
    """
    if valshift == 0:
        mpz_set(out, x)
    elif valshift < 0:
        raise ValueError("negative valuation")
    else:
        mpz_mul(out, x, prime_pow.pow_mpz_t_tmp(valshift))



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/gr.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    slong gr_ctx_sizeof_elem(gr_ctx_t ctx) noexcept
    int gr_ctx_clear(gr_ctx_t ctx) noexcept
    int gr_ctx_write(gr_stream_t out, gr_ctx_t ctx) noexcept
    int gr_ctx_print(gr_ctx_t ctx) noexcept
    int gr_ctx_println(gr_ctx_t ctx) noexcept
    int gr_ctx_get_str(char ** s, gr_ctx_t ctx) noexcept
    int gr_ctx_set_gen_name(gr_ctx_t ctx, const char * s) noexcept
    int gr_ctx_set_gen_names(gr_ctx_t ctx, const char ** s) noexcept
    void gr_init(gr_ptr res, gr_ctx_t ctx) noexcept
    void gr_clear(gr_ptr res, gr_ctx_t ctx) noexcept
    void gr_swap(gr_ptr x, gr_ptr y, gr_ctx_t ctx) noexcept
    void gr_set_shallow(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    gr_ptr gr_heap_init(gr_ctx_t ctx) noexcept
    void gr_heap_clear(gr_ptr x, gr_ctx_t ctx) noexcept
    gr_ptr gr_heap_init_vec(slong len, gr_ctx_t ctx) noexcept
    void gr_heap_clear_vec(gr_ptr x, slong len, gr_ctx_t ctx) noexcept
    int gr_randtest(gr_ptr res, flint_rand_t state, gr_ctx_t ctx) noexcept
    int gr_randtest_not_zero(gr_ptr res, flint_rand_t state, gr_ctx_t ctx) noexcept
    int gr_randtest_small(gr_ptr res, flint_rand_t state, gr_ctx_t ctx) noexcept
    int gr_write(gr_stream_t out, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_print(gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_println(gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_get_str(char ** s, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_set_str(gr_ptr res, const char * x, gr_ctx_t ctx) noexcept
    int gr_write_n(gr_stream_t out, gr_srcptr x, slong n, gr_ctx_t ctx) noexcept
    int gr_get_str_n(char ** s, gr_srcptr x, slong n, gr_ctx_t ctx) noexcept
    int gr_set(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_set_other(gr_ptr res, gr_srcptr x, gr_ctx_t x_ctx, gr_ctx_t ctx) noexcept
    int gr_set_ui(gr_ptr res, ulong x, gr_ctx_t ctx) noexcept
    int gr_set_si(gr_ptr res, slong x, gr_ctx_t ctx) noexcept
    int gr_set_fmpz(gr_ptr res, const fmpz_t x, gr_ctx_t ctx) noexcept
    int gr_set_fmpq(gr_ptr res, const fmpq_t x, gr_ctx_t ctx) noexcept
    int gr_set_d(gr_ptr res, double x, gr_ctx_t ctx) noexcept
    int gr_get_si(slong * res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_get_ui(ulong * res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_get_fmpz(fmpz_t res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_get_fmpq(fmpq_t res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_get_d(double * res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_set_fmpz_2exp_fmpz(gr_ptr res, const fmpz_t x, const fmpz_t y, gr_ctx_t ctx) noexcept
    int gr_get_fmpz_2exp_fmpz(fmpz_t res1, fmpz_t res2, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_get_fexpr(fexpr_t res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_get_fexpr_serialize(fexpr_t res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_set_fexpr(gr_ptr res, fexpr_vec_t inputs, gr_vec_t outputs, const fexpr_t x, gr_ctx_t ctx) noexcept
    int gr_zero(gr_ptr res, gr_ctx_t ctx) noexcept
    int gr_one(gr_ptr res, gr_ctx_t ctx) noexcept
    int gr_neg_one(gr_ptr res, gr_ctx_t ctx) noexcept
    int gr_gen(gr_ptr res, gr_ctx_t ctx) noexcept
    int gr_gens(gr_vec_t res, gr_ctx_t ctx) noexcept
    truth_t gr_is_zero(gr_srcptr x, gr_ctx_t ctx) noexcept
    truth_t gr_is_one(gr_srcptr x, gr_ctx_t ctx) noexcept
    truth_t gr_is_neg_one(gr_srcptr x, gr_ctx_t ctx) noexcept
    truth_t gr_equal(gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    truth_t gr_is_integer(gr_srcptr x, gr_ctx_t ctx) noexcept
    truth_t gr_is_rational(gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_neg(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_add(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_add_ui(gr_ptr res, gr_srcptr x, ulong y, gr_ctx_t ctx) noexcept
    int gr_add_si(gr_ptr res, gr_srcptr x, slong y, gr_ctx_t ctx) noexcept
    int gr_add_fmpz(gr_ptr res, gr_srcptr x, const fmpz_t y, gr_ctx_t ctx) noexcept
    int gr_add_fmpq(gr_ptr res, gr_srcptr x, const fmpq_t y, gr_ctx_t ctx) noexcept
    int gr_add_other(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t y_ctx, gr_ctx_t ctx) noexcept
    int gr_other_add(gr_ptr res, gr_srcptr x, gr_ctx_t x_ctx, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_sub(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_sub_ui(gr_ptr res, gr_srcptr x, ulong y, gr_ctx_t ctx) noexcept
    int gr_sub_si(gr_ptr res, gr_srcptr x, slong y, gr_ctx_t ctx) noexcept
    int gr_sub_fmpz(gr_ptr res, gr_srcptr x, const fmpz_t y, gr_ctx_t ctx) noexcept
    int gr_sub_fmpq(gr_ptr res, gr_srcptr x, const fmpq_t y, gr_ctx_t ctx) noexcept
    int gr_sub_other(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t y_ctx, gr_ctx_t ctx) noexcept
    int gr_other_sub(gr_ptr res, gr_srcptr x, gr_ctx_t x_ctx, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_mul(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_mul_ui(gr_ptr res, gr_srcptr x, ulong y, gr_ctx_t ctx) noexcept
    int gr_mul_si(gr_ptr res, gr_srcptr x, slong y, gr_ctx_t ctx) noexcept
    int gr_mul_fmpz(gr_ptr res, gr_srcptr x, const fmpz_t y, gr_ctx_t ctx) noexcept
    int gr_mul_fmpq(gr_ptr res, gr_srcptr x, const fmpq_t y, gr_ctx_t ctx) noexcept
    int gr_mul_other(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t y_ctx, gr_ctx_t ctx) noexcept
    int gr_other_mul(gr_ptr res, gr_srcptr x, gr_ctx_t x_ctx, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_addmul(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_addmul_ui(gr_ptr res, gr_srcptr x, ulong y, gr_ctx_t ctx) noexcept
    int gr_addmul_si(gr_ptr res, gr_srcptr x, slong y, gr_ctx_t ctx) noexcept
    int gr_addmul_fmpz(gr_ptr res, gr_srcptr x, const fmpz_t y, gr_ctx_t ctx) noexcept
    int gr_addmul_fmpq(gr_ptr res, gr_srcptr x, const fmpq_t y, gr_ctx_t ctx) noexcept
    int gr_addmul_other(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t y_ctx, gr_ctx_t ctx) noexcept
    int gr_submul(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_submul_ui(gr_ptr res, gr_srcptr x, ulong y, gr_ctx_t ctx) noexcept
    int gr_submul_si(gr_ptr res, gr_srcptr x, slong y, gr_ctx_t ctx) noexcept
    int gr_submul_fmpz(gr_ptr res, gr_srcptr x, const fmpz_t y, gr_ctx_t ctx) noexcept
    int gr_submul_fmpq(gr_ptr res, gr_srcptr x, const fmpq_t y, gr_ctx_t ctx) noexcept
    int gr_submul_other(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t y_ctx, gr_ctx_t ctx) noexcept
    int gr_mul_two(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_sqr(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_mul_2exp_si(gr_ptr res, gr_srcptr x, slong y, gr_ctx_t ctx) noexcept
    int gr_mul_2exp_fmpz(gr_ptr res, gr_srcptr x, const fmpz_t y, gr_ctx_t ctx) noexcept
    truth_t gr_is_invertible(gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_inv(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_div(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_div_ui(gr_ptr res, gr_srcptr x, ulong y, gr_ctx_t ctx) noexcept
    int gr_div_si(gr_ptr res, gr_srcptr x, slong y, gr_ctx_t ctx) noexcept
    int gr_div_fmpz(gr_ptr res, gr_srcptr x, const fmpz_t y, gr_ctx_t ctx) noexcept
    int gr_div_fmpq(gr_ptr res, gr_srcptr x, const fmpq_t y, gr_ctx_t ctx) noexcept
    int gr_div_other(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t y_ctx, gr_ctx_t ctx) noexcept
    int gr_other_div(gr_ptr res, gr_srcptr x, gr_ctx_t x_ctx, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_div_nonunique(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    truth_t gr_divides(gr_srcptr d, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_divexact(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_divexact_ui(gr_ptr res, gr_srcptr x, ulong y, gr_ctx_t ctx) noexcept
    int gr_divexact_si(gr_ptr res, gr_srcptr x, slong y, gr_ctx_t ctx) noexcept
    int gr_divexact_fmpz(gr_ptr res, gr_srcptr x, const fmpz_t y, gr_ctx_t ctx) noexcept
    int gr_divexact_other(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t y_ctx, gr_ctx_t ctx) noexcept
    int gr_other_divexact(gr_ptr res, gr_srcptr x, gr_ctx_t x_ctx, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_euclidean_div(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_euclidean_rem(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_euclidean_divrem(gr_ptr res1, gr_ptr res2, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_pow(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_pow_ui(gr_ptr res, gr_srcptr x, ulong y, gr_ctx_t ctx) noexcept
    int gr_pow_si(gr_ptr res, gr_srcptr x, slong y, gr_ctx_t ctx) noexcept
    int gr_pow_fmpz(gr_ptr res, gr_srcptr x, const fmpz_t y, gr_ctx_t ctx) noexcept
    int gr_pow_fmpq(gr_ptr res, gr_srcptr x, const fmpq_t y, gr_ctx_t ctx) noexcept
    int gr_pow_other(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t y_ctx, gr_ctx_t ctx) noexcept
    int gr_other_pow(gr_ptr res, gr_srcptr x, gr_ctx_t x_ctx, gr_srcptr y, gr_ctx_t ctx) noexcept
    truth_t gr_is_square(gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_sqrt(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_rsqrt(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_gcd(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_lcm(gr_ptr res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_factor(gr_ptr c, gr_vec_t factors, gr_vec_t exponents, gr_srcptr x, int flags, gr_ctx_t ctx) noexcept
    int gr_numerator(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_denominator(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_floor(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_ceil(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_trunc(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_nint(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_abs(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_i(gr_ptr res, gr_ctx_t ctx) noexcept
    int gr_conj(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_re(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_im(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_sgn(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_csgn(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_arg(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_pos_inf(gr_ptr res, gr_ctx_t ctx) noexcept
    int gr_neg_inf(gr_ptr res, gr_ctx_t ctx) noexcept
    int gr_uinf(gr_ptr res, gr_ctx_t ctx) noexcept
    int gr_undefined(gr_ptr res, gr_ctx_t ctx) noexcept
    int gr_unknown(gr_ptr res, gr_ctx_t ctx) noexcept
    int gr_cmp(int * res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_cmp_other(int * res, gr_srcptr x, gr_srcptr y, gr_ctx_t y_ctx, gr_ctx_t ctx) noexcept
    int gr_cmpabs(int * res, gr_srcptr x, gr_srcptr y, gr_ctx_t ctx) noexcept
    int gr_cmpabs_other(int * res, gr_srcptr x, gr_srcptr y, gr_ctx_t y_ctx, gr_ctx_t ctx) noexcept
    int gr_ctx_fq_prime(fmpz_t p, gr_ctx_t ctx) noexcept
    int gr_ctx_fq_degree(slong * deg, gr_ctx_t ctx) noexcept
    int gr_ctx_fq_order(fmpz_t q, gr_ctx_t ctx) noexcept
    int gr_fq_frobenius(gr_ptr res, gr_srcptr x, slong e, gr_ctx_t ctx) noexcept
    int gr_fq_multiplicative_order(fmpz_t res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_fq_norm(fmpz_t res, gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_fq_trace(fmpz_t res, gr_srcptr x, gr_ctx_t ctx) noexcept
    truth_t gr_fq_is_primitive(gr_srcptr x, gr_ctx_t ctx) noexcept
    int gr_fq_pth_root(gr_ptr res, gr_srcptr x, gr_ctx_t ctx) noexcept



# sage_setup: distribution = sagemath-modules
# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR
from sage.libs.gsl.types cimport *

cdef extern from "gsl/gsl_sf_gegenbauer.h":

  double  gsl_sf_gegenpoly_1(double lambd, double x)

  double  gsl_sf_gegenpoly_2(double lambd, double x)

  double  gsl_sf_gegenpoly_3(double lambd, double x)

  int  gsl_sf_gegenpoly_1_e(double lambd, double x, gsl_sf_result * result)

  int  gsl_sf_gegenpoly_2_e(double lambd, double x, gsl_sf_result * result)

  int  gsl_sf_gegenpoly_3_e(double lambd, double x, gsl_sf_result * result)

  double  gsl_sf_gegenpoly_n(int n, double lambd, double x)

  int  gsl_sf_gegenpoly_n_e(int n, double lambd, double x, gsl_sf_result * result)

  int  gsl_sf_gegenpoly_array(int nmax, double lambd, double x, double result_array[])




# sage_setup: distribution = sagemath-modules
r"""
Double Precision Real Numbers, implementation using GSL
"""

cimport libc.math

from cysignals.signals cimport sig_on, sig_off

from sage.arith.constants cimport *

from sage.libs.gsl.errno cimport gsl_set_error_handler_off
from sage.libs.gsl.math cimport *
from sage.libs.gsl.exp cimport *
from sage.libs.gsl.log cimport gsl_sf_log
from sage.libs.gsl.trig cimport *
from sage.libs.gsl.dilog cimport gsl_sf_dilog
from sage.libs.gsl.gamma cimport gsl_sf_fact, gsl_sf_gamma
from sage.libs.gsl.zeta cimport gsl_sf_zeta
from sage.libs.gsl.erf cimport gsl_sf_erf


gsl_set_error_handler_off()


cdef class RealDoubleElement_gsl(RealDoubleElement):

    def nth_root(self, int n):
        """
        Return the `n`-th root of ``self``.

        INPUT:

        - ``n`` -- integer

        OUTPUT:

        The output is a complex double if ``self`` is negative and `n` is even,
        otherwise it is a real double.

        EXAMPLES::

            sage: r = RDF(-125.0); r.nth_root(3)
            -5.000000000000001
            sage: r.nth_root(5)
            -2.6265278044037674
            sage: RDF(-2).nth_root(5)^5  # rel tol 1e-15
            -2.000000000000001
            sage: RDF(-1).nth_root(5)^5
            -1.0
            sage: RDF(3).nth_root(10)^10
            2.9999999999999982
            sage: RDF(-1).nth_root(2)
            6.123233995736757e-17 + 1.0*I
            sage: RDF(-1).nth_root(4)
            0.7071067811865476 + 0.7071067811865475*I
        """
        if n == 0:
            return RealDoubleElement(float('nan'))
        if self._value < 0:
            if GSL_IS_EVEN(n):
                from sage.rings.complex_double import CDF
                return self._complex_double_(CDF).nth_root(n)
            else:
                return - ((-self) ** (float(1)/n))

        return self ** (float(1)/n)

    cdef __pow_double(self, double exponent, double sign):
        """
        If ``sign == 1`` or ``self >= 0``, return ``self ^ exponent``.
        If ``sign == -1`` and ``self < 0``, return ``- abs(self) ^ exponent``.
        """
        cdef double v = self._value
        if v >= 0:
            if v == 1:
                return self
            elif exponent == 0:
                return self._new_c(1.0)
            elif v == 0:
                if exponent < 0:
                    raise ZeroDivisionError("0.0 cannot be raised to a negative power")
                return self
            sign = 1.0
        else:  # v < 0
            expmod2 = libc.math.fmod(exponent, 2.0)
            if expmod2 == 0.0:
                pass
            elif expmod2 == 1.0:
                sign = -1.0
            else:
                raise ValueError("negative number cannot be raised to a fractional power")
            v = -v
        return self._new_c(sign * gsl_sf_exp(gsl_sf_log(v) * exponent))

    cpdef _pow_(self, other):
        """
        Return ``self`` raised to the real double power ``other``.

        EXAMPLES::

            sage: a = RDF('1.23456')
            sage: a^a
            1.2971114817819216

        TESTS::

            sage: RDF(0) ^ RDF(0.5)
            0.0
            sage: RDF(0) ^ (1/2)
            0.0
            sage: RDF(0) ^ RDF(0)
            1.0
            sage: RDF(0) ^ RDF(-1)
            Traceback (most recent call last):
            ...
            ZeroDivisionError: 0.0 cannot be raised to a negative power
            sage: RDF(-1) ^ RDF(0)
            1.0
            sage: RDF(-1) ^ RDF(1)
            -1.0
            sage: RDF(-1) ^ RDF(0.5)
            Traceback (most recent call last):
            ...
            ValueError: negative number cannot be raised to a fractional power
        """
        return self.__pow_double((<RealDoubleElement>other)._value, 1)

    cpdef _pow_int(self, n):
        """
        Return ``self`` raised to the integer power ``n``.

        TESTS::

            sage: RDF(1) ^ (2^1000)
            1.0
            sage: RDF(1) ^ (2^1000 + 1)
            1.0
            sage: RDF(1) ^ (-2^1000)
            1.0
            sage: RDF(1) ^ (-2^1000 + 1)
            1.0
            sage: RDF(-1) ^ (2^1000)
            1.0
            sage: RDF(-1) ^ (2^1000 + 1)
            -1.0
            sage: RDF(-1) ^ (-2^1000)
            1.0
            sage: RDF(-1) ^ (-2^1000 + 1)
            -1.0

        ::

            sage: base = RDF(1.0000000000000002)
            sage: base._pow_int(0)
            1.0
            sage: base._pow_int(1)
            1.0000000000000002
            sage: base._pow_int(2)
            1.0000000000000004
            sage: base._pow_int(3)
            1.0000000000000007
            sage: base._pow_int(2^57)
            78962960182680.42
            sage: base._pow_int(2^57 + 1)
            78962960182680.42

        ::

            sage: base = RDF(-1.0000000000000002)
            sage: base._pow_int(0)
            1.0
            sage: base._pow_int(1)
            -1.0000000000000002
            sage: base._pow_int(2)
            1.0000000000000004
            sage: base._pow_int(3)
            -1.0000000000000007
            sage: base._pow_int(2^57)
            78962960182680.42
            sage: base._pow_int(2^57 + 1)
            -78962960182680.42
        """
        return self.__pow_double(n, -1.0 if (n & 1) else 1.0)

    cdef _pow_long(self, long n):
        """
        Compute ``self`` raised to the power ``n``.

        EXAMPLES::

            sage: RDF('1.23456') ^ 20
            67.64629770385...
            sage: RDF(3) ^ 32
            1853020188851841.0
            sage: RDF(2)^(-1024)
            5.562684646268003e-309

        TESTS::

            sage: base = RDF(1.0000000000000002)
            sage: base ^ RDF(2^31)
            1.000000476837272
            sage: base ^ (2^57)
            78962960182680.42
            sage: base ^ RDF(2^57)
            78962960182680.42
        """
        if -2048 <= n <= 2048:
            # For small exponents, it is possible that the powering
            # is exact either because the base is a power of 2
            # (e.g. 2.0^1000) or because the exact result has few
            # significant digits (e.g. 3.0^10). Here, we use the
            # square-and-multiply algorithm by GSL.
            return self._new_c(gsl_pow_int(self._value, <int>n))
        # If the exponent is sufficiently large in absolute value, the
        # result cannot be exact (except if the base is -1.0, 0.0 or
        # 1.0 but those cases are handled by __pow_double too). The
        # log-and-exp algorithm from __pow_double will be more precise
        # than square-and-multiply.

        # We do need to take care of the sign since the conversion
        # of n to double might change an odd number to an even number.
        return self.__pow_double(<double>n, -1.0 if (n & 1) else 1.0)

    cdef _log_base(self, double log_of_base):
        if self._value == 0:
            from sage.rings.real_double import RDF
            return RDF(-1)/RDF(0)
        elif self._value < 0:
            from sage.rings.real_double import RDF
            return RDF.NaN()
        sig_on()
        a = self._new_c(gsl_sf_log(self._value) / log_of_base)
        sig_off()
        return a

    def log(self, base=None):
        """
        Return the logarithm.

        INPUT:

        - ``base`` -- integer or ``None`` (default). The base of the
          logarithm. If ``None`` is specified, the base is `e` (the so-called
          natural logarithm).

        OUTPUT:

        The logarithm of ``self``.  If ``self`` is positive, a double
        floating point number. Infinity if ``self`` is zero. A
        imaginary complex floating point number if ``self`` is
        negative.

        EXAMPLES::

            sage: RDF(2).log()
            0.6931471805599453
            sage: RDF(2).log(2)
            1.0
            sage: RDF(2).log(pi)                                                        # needs sage.symbolic
            0.6055115613982801
            sage: RDF(2).log(10)
            0.30102999566398114
            sage: RDF(2).log(1.5)
            1.7095112913514547
            sage: RDF(0).log()
            -infinity
            sage: RDF(-1).log()
            3.141592653589793*I
            sage: RDF(-1).log(2)  # rel tol 1e-15
            4.532360141827194*I

        TESTS:

        Make sure that we can take the log of small numbers accurately
        and the fix doesn't break preexisting values (:issue:`12557`)::

            sage: R = RealField(128)
            sage: def check_error(x):
            ....:   x = RDF(x)
            ....:   log_RDF = x.log()
            ....:   log_RR = R(x).log()
            ....:   diff = R(log_RDF) - log_RR
            ....:   if abs(diff) < log_RDF.ulp():
            ....:       return True
            ....:   print("logarithm check failed for %s (diff = %s ulp)"% \
            ....:       (x, diff/log_RDF.ulp()))
            ....:   return False
            sage: all( check_error(2^x) for x in range(-100,100) )
            True
            sage: all( check_error(x) for x in sxrange(0.01, 2.00, 0.01) )
            True
            sage: all( check_error(x) for x in sxrange(0.99, 1.01, 0.001) )
            True
            sage: RDF(1.000000001).log()
            1.000000082240371e-09
            sage: RDF(1e-17).log()
            -39.14394658089878
            sage: RDF(1e-50).log()
            -115.12925464970229
        """
        if self < 0:
            from sage.rings.complex_double import CDF
            return CDF(self).log(base)
        if base is None:
            return self._log_base(1)
        else:
            if isinstance(base, RealDoubleElement):
                return self._log_base(base._log_base(1))
            else:
                return self._log_base(gsl_sf_log(float(base)))

    def log2(self):
        """
        Return log to the base 2 of ``self``.

        EXAMPLES::

            sage: r = RDF(16.0)
            sage: r.log2()
            4.0

        ::

            sage: r = RDF(31.9); r.log2()
            4.9954845188775066
        """
        if self < 0:
            from sage.rings.complex_double import CDF
            return CDF(self).log(2)
        sig_on()
        a = self._new_c(gsl_sf_log(self._value) * M_1_LN2)
        sig_off()
        return a

    def log10(self):
        """
        Return log to the base 10 of ``self``.

        EXAMPLES::

            sage: r = RDF('16.0'); r.log10()
            1.2041199826559248
            sage: r.log() / RDF(log(10))                                                # needs sage.symbolic
            1.2041199826559246
            sage: r = RDF('39.9'); r.log10()
            1.6009728956867482
        """
        if self < 0:
            from sage.rings.complex_double import CDF
            return CDF(self).log(10)
        sig_on()
        a = self._new_c(gsl_sf_log(self._value) * M_1_LN10)
        sig_off()
        return a

    def logpi(self):
        r"""
        Return log to the base `\pi` of ``self``.

        EXAMPLES::

            sage: r = RDF(16); r.logpi()
            2.4220462455931204
            sage: r.log() / RDF(log(pi))                                                # needs sage.symbolic
            2.4220462455931204
            sage: r = RDF('39.9'); r.logpi()
            3.2203023346075152
        """
        if self < 0:
            from sage.rings.complex_double import CDF
            return CDF(self).log(M_PI)
        sig_on()
        a = self._new_c(gsl_sf_log(self._value) * M_1_LNPI)
        sig_off()
        return a

    def exp(self):
        r"""
        Return `e^\mathtt{self}`.

        EXAMPLES::

            sage: r = RDF(0.0)
            sage: r.exp()
            1.0

        ::

            sage: r = RDF('32.3')
            sage: a = r.exp(); a
            106588847274864.47
            sage: a.log()
            32.3

        ::

            sage: r = RDF('-32.3')
            sage: r.exp()
            9.381844588498685e-15

        ::

            sage: RDF(1000).exp()
            +infinity
        """
        sig_on()
        a = self._new_c(gsl_sf_exp(self._value))
        sig_off()
        return a

    def exp2(self):
        r"""
        Return `2^\mathtt{self}`.

        EXAMPLES::

            sage: r = RDF(0.0)
            sage: r.exp2()
            1.0

        ::

            sage: r = RDF(32.0)
            sage: r.exp2()
            4294967295.9999967

        ::

            sage: r = RDF(-32.3)
            sage: r.exp2()
            1.8911724825302065e-10
        """
        sig_on()
        a = self._new_c(gsl_sf_exp(self._value * M_LN2))
        sig_off()
        return a

    def exp10(self):
        r"""
        Return `10^\mathtt{self}`.

        EXAMPLES::

            sage: r = RDF(0.0)
            sage: r.exp10()
            1.0

        ::

            sage: r = RDF(32.0)
            sage: r.exp10()
            1.0000000000000069e+32

        ::

            sage: r = RDF(-32.3)
            sage: r.exp10()
            5.011872336272702e-33
        """
        sig_on()
        a = self._new_c(gsl_sf_exp(self._value * M_LN10))
        sig_off()
        return a

    def cos(self):
        """
        Return the cosine of ``self``.

        EXAMPLES::

            sage: t=RDF.pi()/2
            sage: t.cos()
            6.123233995736757e-17
        """
        return self._new_c(gsl_sf_cos(self._value))

    def sin(self):
        """
        Return the sine of ``self``.

        EXAMPLES::

            sage: RDF(2).sin()
            0.9092974268256817
        """
        return self._new_c(gsl_sf_sin(self._value))

    def dilog(self):
        r"""
        Return the dilogarithm of ``self``.

        This is defined by the
        series `\sum_n x^n/n^2` for `|x| \le 1`. When the absolute
        value of ``self`` is greater than 1, the returned value is the
        real part of (the analytic continuation to `\CC` of) the
        dilogarithm of ``self``.

        EXAMPLES::

            sage: RDF(1).dilog()  # rel tol 1.0e-13
            1.6449340668482264
            sage: RDF(2).dilog()  # rel tol 1.0e-13
            2.46740110027234
        """
        return self._new_c(gsl_sf_dilog(self._value))

    def restrict_angle(self):
        r"""
        Return a number congruent to ``self`` mod `2\pi` that lies in
        the interval `(-\pi, \pi]`.

        Specifically, it is the unique `x \in (-\pi, \pi]` such
        that ``self`` `= x + 2\pi n` for some `n \in \ZZ`.

        EXAMPLES::

            sage: RDF(pi).restrict_angle()                                              # needs sage.symbolic
            3.141592653589793
            sage: RDF(pi + 1e-10).restrict_angle()                                      # needs sage.symbolic
            -3.1415926534897936
            sage: RDF(1+10^10*pi).restrict_angle()                                      # needs sage.symbolic
            0.9999977606...
        """
        return self._new_c(gsl_sf_angle_restrict_symm(self._value))

    def tan(self):
        """
        Return the tangent of ``self``.

        EXAMPLES::

            sage: q = RDF.pi()/3
            sage: q.tan()
            1.7320508075688767
            sage: q = RDF.pi()/6
            sage: q.tan()
            0.5773502691896256
        """
        cos = gsl_sf_cos(self._value)
        a = self._new_c(gsl_sf_sin(self._value) / cos)
        return a

    def sincos(self):
        """
        Return a pair consisting of the sine and cosine of ``self``.

        EXAMPLES::

            sage: t = RDF.pi()/6
            sage: t.sincos()
            (0.49999999999999994, 0.8660254037844387)
        """
        return self.sin(), self.cos()

    def hypot(self, other):
        r"""
        Compute the value `\sqrt{s^2 + o^2}` where `s` is ``self`` and `o`
        is ``other`` in such a way as to avoid overflow.

        EXAMPLES::

            sage: x = RDF(4e300); y = RDF(3e300)
            sage: x.hypot(y)
            5e+300
            sage: sqrt(x^2+y^2) # overflow
            +infinity
        """
        sig_on()
        a = self._new_c(gsl_sf_hypot(self._value, float(other)))
        sig_off()
        return a

    def arccos(self):
        """
        Return the inverse cosine of ``self``.

        EXAMPLES::

            sage: q = RDF.pi()/3
            sage: i = q.cos()
            sage: i.arccos() == q
            True
        """
        return self._new_c(libc.math.acos(self._value))

    def arcsin(self):
        """
        Return the inverse sine of ``self``.

        EXAMPLES::

            sage: q = RDF.pi()/5
            sage: i = q.sin()
            sage: i.arcsin() == q
            True
        """
        return self._new_c(libc.math.asin(self._value))

    def arctan(self):
        """
        Return the inverse tangent of ``self``.

        EXAMPLES::

            sage: q = RDF.pi()/5
            sage: i = q.tan()
            sage: i.arctan() == q
            True
        """
        return self._new_c(libc.math.atan(self._value))

    def cosh(self):
        """
        Return the hyperbolic cosine of ``self``.

        EXAMPLES::

            sage: q = RDF.pi()/12
            sage: q.cosh()
            1.0344656400955106
        """
        return self._new_c(gsl_ldexp(gsl_sf_exp(self._value) + gsl_sf_exp(-self._value), -1))  # (e^x + e^-x)/2

    def sinh(self):
        """
        Return the hyperbolic sine of ``self``.

        EXAMPLES::

            sage: q = RDF.pi()/12
            sage: q.sinh()
            0.26480022760227073
        """
        return self._new_c(gsl_ldexp(gsl_sf_expm1(self._value) - gsl_sf_expm1(-self._value), -1))  # (e^x - e^-x)/2

    def tanh(self):
        """
        Return the hyperbolic tangent of ``self``.

        EXAMPLES::

            sage: q = RDF.pi()/12
            sage: q.tanh()
            0.25597778924568454
        """
        return self.sinh() / self.cosh()

    def acosh(self):
        """
        Return the hyperbolic inverse cosine of ``self``.

        EXAMPLES::

            sage: q = RDF.pi()/2
            sage: i = q.cosh(); i
            2.5091784786580567
            sage: abs(i.acosh()-q) < 1e-15
            True
        """
        return self._new_c(gsl_acosh(self._value))

    def arcsinh(self):
        """
        Return the hyperbolic inverse sine of ``self``.

        EXAMPLES::

            sage: q = RDF.pi()/2
            sage: i = q.sinh(); i
            2.3012989023072947
            sage: abs(i.arcsinh()-q) < 1e-15
            True
        """
        return self._new_c(gsl_asinh(self._value))

    def arctanh(self):
        """
        Return the hyperbolic inverse tangent of ``self``.

        EXAMPLES::

            sage: q = RDF.pi()/2
            sage: i = q.tanh(); i
            0.9171523356672744
            sage: i.arctanh() - q  # rel tol 1
            4.440892098500626e-16
        """
        return self._new_c(gsl_atanh(self._value))

    def sech(self):
        r"""
        Return the hyperbolic secant of ``self``.

        EXAMPLES::

            sage: RDF(pi).sech()                                                        # needs sage.symbolic
            0.08626673833405443
            sage: CDF(pi).sech()                                                        # needs sage.symbolic
            0.08626673833405443
        """
        return 1/self.cosh()

    def csch(self):
        r"""
        Return the hyperbolic cosecant of ``self``.

        EXAMPLES::

            sage: RDF(pi).csch()                                                        # needs sage.symbolic
            0.08658953753004694
            sage: CDF(pi).csch()  # rel tol 1e-15                                       # needs sage.symbolic
            0.08658953753004696
        """
        return 1/self.sinh()

    def coth(self):
        r"""
        Return the hyperbolic cotangent of ``self``.

        EXAMPLES::

            sage: RDF(pi).coth()                                                        # needs sage.symbolic
            1.003741873197321
            sage: CDF(pi).coth()                                                        # needs sage.symbolic
            1.0037418731973213
        """
        return self.cosh() / self.sinh()

    def erf(self):
        """
        Return the value of the error function on ``self``.

        EXAMPLES::

            sage: RDF(6).erf()
            1.0
        """
        return self._new_c(gsl_sf_erf(self._value))

    @classmethod
    def _factorial(cls, int n):
        """
        Return the factorial of the integer `n` as a real number.

        EXAMPLES::

            sage: RDF.factorial(100)
            9.332621544394415e+157
        """
        if n < 0:
            raise ArithmeticError("n must be nonnegative")
        return cls(gsl_sf_fact(n))

    def gamma(self):
        """
        Return the value of the Euler gamma function on ``self``.

        EXAMPLES::

            sage: RDF(6).gamma()
            120.0
            sage: RDF(1.5).gamma()  # rel tol 1e-15
            0.8862269254527584
        """
        sig_on()
        a = self._new_c(gsl_sf_gamma(self._value))
        sig_off()
        return a

    def zeta(self):
        r"""
        Return the Riemann zeta function evaluated at this real number.

        .. NOTE::

           PARI is vastly more efficient at computing the Riemann zeta
           function. See the example below for how to use it.

        EXAMPLES::

            sage: RDF(2).zeta()  # rel tol 1e-15
            1.6449340668482269
            sage: RDF.pi()^2/6
            1.6449340668482264
            sage: RDF(-2).zeta()
            0.0
            sage: RDF(1).zeta()
            +infinity
        """
        if self._value == 1:
            return self._new_c(1)/self._new_c(0)
        return self._new_c(gsl_sf_zeta(self._value))



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/bool_mat.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    int bool_mat_get_entry(const bool_mat_t mat, slong i, slong j) noexcept
    void bool_mat_set_entry(bool_mat_t mat, slong i, slong j, int x) noexcept
    void bool_mat_init(bool_mat_t mat, slong r, slong c) noexcept
    void bool_mat_clear(bool_mat_t mat) noexcept
    bint bool_mat_is_empty(const bool_mat_t mat) noexcept
    bint bool_mat_is_square(const bool_mat_t mat) noexcept
    void bool_mat_set(bool_mat_t dest, const bool_mat_t src) noexcept
    void bool_mat_print(const bool_mat_t mat) noexcept
    void bool_mat_fprint(FILE * file, const bool_mat_t mat) noexcept
    bint bool_mat_equal(const bool_mat_t mat1, const bool_mat_t mat2) noexcept
    int bool_mat_any(const bool_mat_t mat) noexcept
    int bool_mat_all(const bool_mat_t mat) noexcept
    bint bool_mat_is_diagonal(const bool_mat_t A) noexcept
    bint bool_mat_is_lower_triangular(const bool_mat_t A) noexcept
    bint bool_mat_is_transitive(const bool_mat_t mat) noexcept
    bint bool_mat_is_nilpotent(const bool_mat_t A) noexcept
    void bool_mat_randtest(bool_mat_t mat, flint_rand_t state) noexcept
    void bool_mat_randtest_diagonal(bool_mat_t mat, flint_rand_t state) noexcept
    void bool_mat_randtest_nilpotent(bool_mat_t mat, flint_rand_t state) noexcept
    void bool_mat_zero(bool_mat_t mat) noexcept
    void bool_mat_one(bool_mat_t mat) noexcept
    void bool_mat_directed_path(bool_mat_t A) noexcept
    void bool_mat_directed_cycle(bool_mat_t A) noexcept
    void bool_mat_transpose(bool_mat_t dest, const bool_mat_t src) noexcept
    void bool_mat_complement(bool_mat_t B, const bool_mat_t A) noexcept
    void bool_mat_add(bool_mat_t res, const bool_mat_t mat1, const bool_mat_t mat2) noexcept
    void bool_mat_mul(bool_mat_t res, const bool_mat_t mat1, const bool_mat_t mat2) noexcept
    void bool_mat_mul_entrywise(bool_mat_t res, const bool_mat_t mat1, const bool_mat_t mat2) noexcept
    void bool_mat_sqr(bool_mat_t B, const bool_mat_t A) noexcept
    void bool_mat_pow_ui(bool_mat_t B, const bool_mat_t A, ulong exp) noexcept
    int bool_mat_trace(const bool_mat_t mat) noexcept
    slong bool_mat_nilpotency_degree(const bool_mat_t A) noexcept
    void bool_mat_transitive_closure(bool_mat_t B, const bool_mat_t A) noexcept
    slong bool_mat_get_strongly_connected_components(slong * p, const bool_mat_t A) noexcept
    slong bool_mat_all_pairs_longest_walk(fmpz_mat_t B, const bool_mat_t A) noexcept



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.libs.singular
r"""
Free resolutions

Let `R` be a commutative ring. A finite free resolution of an `R`-module `M`
is a chain complex of free `R`-modules

.. MATH::

    0 \rightarrow R^{n_k} \xrightarrow{d_k}
    \cdots \xrightarrow{d_2} R^{n_1} \xrightarrow{d_1} R^{n_0}

terminating with a zero module at the end that is exact (all homology groups
are zero) such that the image of `d_1` is `M`.

EXAMPLES::

    sage: from sage.homology.free_resolution import FreeResolution
    sage: S.<x,y,z,w> = PolynomialRing(QQ)
    sage: m = matrix(S, 1, [z^2 - y*w, y*z - x*w, y^2 - x*z]).transpose()
    sage: r = FreeResolution(m, name='S'); r
    S^1 <-- S^3 <-- S^2 <-- 0

    sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
    sage: r = I.free_resolution(); r
    S^1 <-- S^3 <-- S^2 <-- 0

::

    sage: S.<x,y,z,w> = PolynomialRing(QQ)
    sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
    sage: r = I.graded_free_resolution(); r
    S(0) <-- S(-2)⊕S(-2)⊕S(-2) <-- S(-3)⊕S(-3) <-- 0

An example of a minimal free resolution from [CLO2005]_::

    sage: R.<x,y,z,w> = QQ[]
    sage: I = R.ideal([y*z - x*w, y^3 - x^2*z, x*z^2 - y^2*w, z^3 - y*w^2])
    sage: r = I.free_resolution();  r
    S^1 <-- S^4 <-- S^4 <-- S^1 <-- 0
    sage: len(r)
    3
    sage: r.matrix(2)
    [-z^2 -x*z  y*w -y^2]
    [   y    0   -x    0]
    [  -w    y    z    x]
    [   0    w    0    z]

AUTHORS:

- Kwankyu Lee (2022-05-13): initial version
- Travis Scrimshaw (2022-08-23): refactored for free module inputs
"""

# ****************************************************************************
#       Copyright (C) 2022 Kwankyu Lee <ekwankyu@gmail.com>
#                 (C) 2022 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.lazy_attribute import lazy_attribute
from sage.misc.abstract_method import abstract_method
from sage.misc.classcall_metaclass import ClasscallMetaclass
from sage.structure.sage_object import SageObject
from sage.structure.element import Matrix
from sage.categories.principal_ideal_domains import PrincipalIdealDomains
from sage.categories.integral_domains import IntegralDomains
from sage.modules.free_module_element import vector
from sage.modules.free_module import FreeModule
from sage.modules.free_module import Module_free_ambient, FreeModule_generic
from sage.rings.ideal import Ideal_generic

from copy import copy


class FreeResolution(SageObject, metaclass=ClasscallMetaclass):
    r"""
    A free resolution.

    Let `R` be a commutative ring. A *free resolution* of an `R`-module `M`
    is a (possibly infinite) chain complex of free `R`-modules

    .. MATH::

        \cdots \rightarrow R^{n_k} \xrightarrow{d_k}
        \cdots \xrightarrow{d_2} R^{n_1} \xrightarrow{d_1} R^{n_0}

    that is exact (all homology groups are zero) such that the image
    of `d_1` is `M`.
    """
    @staticmethod
    def __classcall_private__(cls, module, *args, graded=False, degrees=None, shifts=None, **kwds):
        """
        Dispatch to the correct constructor.

        TESTS::

            sage: from sage.homology.free_resolution import FreeResolution
            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: m = matrix(S, 1, [z^2 - y*w, y*z - x*w, y^2 - x*z]).transpose()
            sage: r = FreeResolution(m, name='S')
            sage: type(r)
            <class 'sage.homology.free_resolution.FiniteFreeResolution_singular'>

            sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
            sage: r = FreeResolution(I)
            sage: type(r)
            <class 'sage.homology.free_resolution.FiniteFreeResolution_singular'>

            sage: R.<x> = QQ[]
            sage: M = R^3
            sage: v = M([x^2, 2*x^2, 3*x^2])
            sage: w = M([0, x, 2*x])
            sage: S = M.submodule([v, w])
            sage: r = FreeResolution(S)
            sage: type(r)
            <class 'sage.homology.free_resolution.FiniteFreeResolution_free_module'>

            sage: I = R.ideal([x^4 + 3*x^2 + 2])
            sage: r = FreeResolution(I)
            sage: type(r)
            <class 'sage.homology.free_resolution.FiniteFreeResolution_free_module'>

            sage: R.<x,y> = QQ[]
            sage: I = R.ideal([x^2, y^3])
            sage: Q = R.quo(I)
            sage: Q.is_integral_domain()
            False
            sage: xb, yb = Q.gens()                                                     # needs sage.rings.function_field
            sage: FreeResolution(Q.ideal([xb]))  # has torsion                          # needs sage.rings.function_field
            Traceback (most recent call last):
            ...
            NotImplementedError: the ring must be a polynomial ring using Singular
        """
        if degrees is not None or shifts is not None:
            graded = True

        if isinstance(module, Matrix):
            is_free_module = False
            S = module.base_ring()
            if S in PrincipalIdealDomains():
                module = module.echelon_form()
                if module.nrows() > module.rank():
                    module = module.submatrix(nrows=module.rank())
                    module.set_immutable()
                is_free_module = True
            if not module.is_immutable():
                # We need to make an immutable copy of the matrix
                module = copy(module)
                module.set_immutable()
            if is_free_module:
                if graded:
                    from sage.homology.graded_resolution import GradedFiniteFreeResolution_free_module
                    return GradedFiniteFreeResolution_free_module(module,
                                                                  *args,
                                                                  degrees=degrees,
                                                                  shifts=shifts,
                                                                  **kwds)
                return FiniteFreeResolution_free_module(module, *args, **kwds)

            from sage.rings.polynomial.multi_polynomial_libsingular import MPolynomialRing_libsingular
            if not isinstance(S, MPolynomialRing_libsingular):
                raise NotImplementedError("the matrix must be over a PID or a "
                                          " polynomial ring that is using Singular")

            if graded:
                # We are computing a graded resolution
                from sage.homology.graded_resolution import GradedFiniteFreeResolution_singular
                return GradedFiniteFreeResolution_singular(module, *args, degrees=degrees,
                                                           shifts=shifts, **kwds)

            return FiniteFreeResolution_singular(module, **kwds)

        if graded:
            return module.graded_free_resolution(*args, **kwds)
        return module.free_resolution(*args, **kwds)

    def __init__(self, module, name='S', **kwds) -> None:
        """
        Initialize ``self``.

        INPUT:

        - ``base_ring`` -- a ring
        - ``name`` -- (default: ``'S'``) the name of the ring for printing

        TESTS::

            sage: from sage.homology.free_resolution import FreeResolution
            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: m1 = matrix(S, 1, [z^2 - y*w, y*z - x*w, y^2 - x*z])
            sage: r = FreeResolution(m1, name='S')
            sage: TestSuite(r).run(skip=['_test_pickling'])
        """
        if isinstance(module, Ideal_generic):
            S = module.ring()
        else:  # module or matrix
            S = module.base_ring()

        self._base_ring = S
        self._name = name
        self._module = module

    def _repr_(self) -> str:
        r"""
        Return a string representation of ``self``.

        TESTS::

            sage: from sage.homology.free_resolution import FreeResolution
            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: m1 = matrix(S, 1, [z^2 - y*w, y*z - x*w, y^2 - x*z])
            sage: r = FreeResolution(m1, name='S')
            sage: print(FreeResolution._repr_(r))
            Free resolution of the row space of the matrix:
            [z^2 - y*w y*z - x*w y^2 - x*z]
        """
        if isinstance(self._module, Matrix):
            return f"Free resolution of the row space of the matrix:\n{self._module}"
        return f"Free resolution of {self._module}"

    def _repr_module(self, i) -> str:
        r"""
        Return the string form of the `i`-th free module.

        INPUT:

        - ``i`` -- positive integer

        EXAMPLES::

            sage: from sage.homology.free_resolution import FreeResolution
            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: m = matrix(S, 1, [y*w - z^2, -x*w + y*z, x*z - y^2])
            sage: r = FreeResolution(m.transpose(), name='S')
            sage: r._repr_module(2)
            'S^2'
            sage: r  # indirect doctest
            S^1 <-- S^3 <-- S^2 <-- 0

        TESTS::

            sage: S.<x,y,z> = PolynomialRing(QQ)
            sage: I = S.ideal(0)
            sage: C = I.free_resolution()
            sage: C
            S^1 <-- 0
        """
        if i == 0:
            if self._length > 0:
                r = self._maps[0].nrows()
            else:
                r = self._initial_differential.domain().dimension()
            s = f'{self._name}^{r}'
            return s
        elif i > self._length:
            s = '0'
        else:
            r = self._maps[i - 1].ncols()
            if r > 0:
                s = f'{self._name}^{r}'
            else:
                s = '0'
        return s

    @abstract_method
    def differential(self, i):
        r"""
        Return the `i`-th differential map.

        INPUT:

        - ``i`` -- positive integer

        TESTS::

            sage: from sage.homology.free_resolution import FreeResolution
            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: m1 = matrix(S, 1, [z^2 - y*w, y*z - x*w, y^2 - x*z])
            sage: r = FreeResolution(m1, name='S')
            sage: FreeResolution.differiental(r, 1)
            Traceback (most recent call last):
            ...
            AttributeError: type object 'FreeResolution' has no attribute 'differiental'...
        """

    def target(self):
        r"""
        Return the codomain of the `0`-th differential map.

        The codomain of the `0`-th differential map is the cokernel of
        the first differential map.

        EXAMPLES::

            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
            sage: r = I.graded_free_resolution()
            sage: r
            S(0) <-- S(-2)⊕S(-2)⊕S(-2) <-- S(-3)⊕S(-3) <-- 0
            sage: r.target()
            Quotient module by
             Submodule of Ambient free module of rank 1 over the integral domain
              Multivariate Polynomial Ring in x, y, z, w over Rational Field
              Generated by the rows of the matrix:
              [-z^2 + y*w]
              [ y*z - x*w]
              [-y^2 + x*z]
        """
        return self.differential(0).codomain()


class FiniteFreeResolution(FreeResolution):
    r"""
    Finite free resolutions.

    The matrix at index `i` in the list defines the differential map from
    `(i + 1)`-th free module to the `i`-th free module over the base ring by
    multiplication on the left. The number of matrices in the list is the
    length of the resolution. The number of rows and columns of the matrices
    define the ranks of the free modules in the resolution.

    Note that the first matrix in the list defines the differential map at
    homological index `1`.

    A subclass must provide a ``_maps`` attribute that contains a list of the
    maps defining the resolution.

    A subclass can define ``_initial_differential`` attribute that
    contains the `0`-th differential map whose codomain is the target
    of the free resolution.

    EXAMPLES::

        sage: from sage.homology.free_resolution import FreeResolution
        sage: S.<x,y,z,w> = PolynomialRing(QQ)
        sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
        sage: r = FreeResolution(I)
        sage: r.differential(0)
        Coercion map:
          From: Ambient free module of rank 1 over the integral domain
                Multivariate Polynomial Ring in x, y, z, w over Rational Field
          To:   Quotient module by
                Submodule of Ambient free module of rank 1 over the integral domain
                Multivariate Polynomial Ring in x, y, z, w over Rational Field
                Generated by the rows of the matrix:
                [-z^2 + y*w]
                [ y*z - x*w]
                [-y^2 + x*z]
    """
    @lazy_attribute
    def _length(self):
        """
        The length of ``self``.

        TESTS::

            sage: from sage.homology.free_resolution import FreeResolution
            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
            sage: r = FreeResolution(I)
            sage: r._length
            2
        """
        return len(self._maps)

    def _repr_(self) -> str:
        """
        Return the string form of this resolution.

        INPUT:

        - ``i`` -- positive integer

        EXAMPLES::

            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
            sage: r = I.graded_free_resolution(); r
            S(0) <-- S(-2)⊕S(-2)⊕S(-2) <-- S(-3)⊕S(-3) <-- 0
        """
        s = self._repr_module(0)
        for i in range(1, self._length + 1):
            s += ' <-- ' + self._repr_module(i)
        s += ' <-- 0'
        return s

    def __len__(self):
        r"""
        Return the length of this resolution.

        The length of a free resolution is the index of the last nonzero free module.

        EXAMPLES::

            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
            sage: r = I.graded_free_resolution(); r
            S(0) <-- S(-2)⊕S(-2)⊕S(-2) <-- S(-3)⊕S(-3) <-- 0
            sage: len(r)
            2
        """
        return len(self._maps)

    def __getitem__(self, i):
        r"""
        Return the `i`-th free module of this resolution.

        INPUT:

        - ``i`` -- positive integer

        EXAMPLES::

            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
            sage: r = I.graded_free_resolution(); r
            S(0) <-- S(-2)⊕S(-2)⊕S(-2) <-- S(-3)⊕S(-3) <-- 0
            sage: r.target()
            Quotient module by Submodule of Ambient free module of rank 1 over the integral domain
            Multivariate Polynomial Ring in x, y, z, w over Rational Field
            Generated by the rows of the matrix:
            [-z^2 + y*w]
            [ y*z - x*w]
            [-y^2 + x*z]
        """
        if i < 0:
            raise IndexError('invalid index')
        elif i > self._length:
            F = FreeModule(self._base_ring, 0)
        elif i == 0:
            F = self.differential(0).domain()
        elif i == self._length:
            F = FreeModule(self._base_ring, self._maps[i - 1].ncols())
        else:
            F = FreeModule(self._base_ring, self._maps[i].nrows())
        return F

    def differential(self, i):
        r"""
        Return the `i`-th differential map.

        INPUT:

        - ``i`` -- positive integer

        EXAMPLES::

            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
            sage: r = I.graded_free_resolution()
            sage: r
            S(0) <-- S(-2)⊕S(-2)⊕S(-2) <-- S(-3)⊕S(-3) <-- 0
            sage: r.differential(3)
            Free module morphism defined as left-multiplication by the matrix
             []
             Domain:   Ambient free module of rank 0 over the integral domain
                       Multivariate Polynomial Ring in x, y, z, w over Rational Field
             Codomain: Ambient free module of rank 2 over the integral domain
                       Multivariate Polynomial Ring in x, y, z, w over Rational Field
            sage: r.differential(2)
            Free module morphism defined as left-multiplication by the matrix
              [-y  x]
              [ z -y]
              [-w  z]
              Domain:   Ambient free module of rank 2 over the integral domain
                        Multivariate Polynomial Ring in x, y, z, w over Rational Field
              Codomain: Ambient free module of rank 3 over the integral domain
                        Multivariate Polynomial Ring in x, y, z, w over Rational Field
            sage: r.differential(1)
            Free module morphism defined as left-multiplication by the matrix
              [z^2 - y*w y*z - x*w y^2 - x*z]
              Domain:   Ambient free module of rank 3 over the integral domain
                        Multivariate Polynomial Ring in x, y, z, w over Rational Field
              Codomain: Ambient free module of rank 1 over the integral domain
                        Multivariate Polynomial Ring in x, y, z, w over Rational Field
            sage: r.differential(0)
            Coercion map:
              From: Ambient free module of rank 1 over the integral domain
                    Multivariate Polynomial Ring in x, y, z, w over Rational Field
              To:   Quotient module by
                    Submodule of Ambient free module of rank 1 over the integral domain
                    Multivariate Polynomial Ring in x, y, z, w over Rational Field
                    Generated by the rows of the matrix:
                    [-z^2 + y*w]
                    [ y*z - x*w]
                    [-y^2 + x*z]

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: S = P2.coordinate_ring()
            sage: I = S.ideal(0)
            sage: C = I.graded_free_resolution(); C
            S(0) <-- 0
            sage: C[1]
            Ambient free module of rank 0 over the integral domain
             Multivariate Polynomial Ring in x, y, z over Rational Field
            sage: C[0]
            Ambient free module of rank 1 over the integral domain
             Multivariate Polynomial Ring in x, y, z over Rational Field
            sage: C.differential(1)
            Free module morphism defined as left-multiplication by the matrix
            []
            Domain: Ambient free module of rank 0 over the integral domain
             Multivariate Polynomial Ring in x, y, z over Rational Field
            Codomain: Ambient free module of rank 1 over the integral domain
             Multivariate Polynomial Ring in x, y, z over Rational Field
            sage: C.differential(1).matrix()
            []
            sage: C.differential(1).matrix().dimensions()
            (1, 0)
        """
        if i < 0:
            raise IndexError('invalid index')
        elif i == 0:
            try:
                return self._initial_differential
            except AttributeError:
                raise ValueError('0th differential map undefined')
        elif i > self._length + 1:
            s = FreeModule(self._base_ring, 0)
            t = FreeModule(self._base_ring, 0)
            m = s.hom(0, t, side='right')
        elif i == self._length + 1:
            s = FreeModule(self._base_ring, 0)
            if self._length > 0:
                t = FreeModule(self._base_ring, self._maps[i - 2].ncols())
            else:
                t = self._initial_differential.domain()
            m = s.hom(0, t, side='right')
        else:
            s = FreeModule(self._base_ring, self._maps[i - 1].ncols())
            t = FreeModule(self._base_ring, self._maps[i - 1].nrows())
            m = s.hom(self._maps[i - 1], t, side='right')
        return m

    def matrix(self, i):
        r"""
        Return the matrix representing the `i`-th differential map.

        INPUT:

        - ``i`` -- positive integer

        EXAMPLES::

            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
            sage: r = I.graded_free_resolution(); r
            S(0) <-- S(-2)⊕S(-2)⊕S(-2) <-- S(-3)⊕S(-3) <-- 0
            sage: r.matrix(3)
            []
            sage: r.matrix(2)
            [-y  x]
            [ z -y]
            [-w  z]
            sage: r.matrix(1)
            [z^2 - y*w y*z - x*w y^2 - x*z]
        """
        if i <= 0:
            raise IndexError('invalid index')
        elif i <= self._length:
            return self._maps[i - 1]
        else:
            return self.differential(i).matrix()

    def chain_complex(self):
        r"""
        Return this resolution as a chain complex.

        A chain complex in Sage has its own useful methods.

        EXAMPLES::

            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
            sage: r = I.graded_free_resolution()
            sage: unicode_art(r.chain_complex())
                                                           ⎛-y  x⎞
                                                           ⎜ z -y⎟
                       (z^2 - y*w y*z - x*w y^2 - x*z)     ⎝-w  z⎠
             0 <── C_0 <────────────────────────────── C_1 <────── C_2 <── 0
        """
        from sage.homology.chain_complex import ChainComplex
        mats = {}
        for i in range(self._length, 0, -1):
            mats[i] = self.matrix(i)
        return ChainComplex(mats, degree_of_differential=-1)

    @lazy_attribute
    def _initial_differential(self):
        r"""
        Define the `0`-th differential map of this resolution.

        EXAMPLES::

            sage: from sage.homology.free_resolution import FreeResolution
            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
            sage: r = FreeResolution(I)
            sage: r._initial_differential
            Coercion map:
              From: Ambient free module of rank 1 over the integral domain
                    Multivariate Polynomial Ring in x, y, z, w over Rational Field
              To:   Quotient module by
                    Submodule of Ambient free module of rank 1 over the integral domain
                    Multivariate Polynomial Ring in x, y, z, w over Rational Field
            Generated by the rows of the matrix:
            [-z^2 + y*w]
            [ y*z - x*w]
            [-y^2 + x*z]
        """
        module = self._module
        if isinstance(module, Ideal_generic):
            S = module.ring()
            M = FreeModule(S, 1)
            N = M.submodule([vector([g]) for g in module.gens()])
        elif isinstance(module, Module_free_ambient):
            S = module.base_ring()
            M = module.ambient_module()
            N = module
        elif isinstance(module, Matrix):
            S = module.base_ring()
            N = module.row_space()
            M = N.ambient_module()
        Q = M.quotient(N)
        return Q.coerce_map_from(M)

    def _m(self):
        r"""
        Return the matrix whose column space is ``self._module``.

        If ``self._module`` is an ideal, then just the ideal is returned.

        TESTS::

            sage: from sage.homology.free_resolution import FreeResolution
            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
            sage: r = FreeResolution(I)
            sage: r._m()
            Ideal (-z^2 + y*w, y*z - x*w, -y^2 + x*z) of
             Multivariate Polynomial Ring in x, y, z, w over Rational Field

            sage: m = matrix(S, 1, [z^2 - y*w, y*z - x*w, y^2 - x*z]).transpose()
            sage: r = FreeResolution(m, name='S')
            sage: r._m()
            [z^2 - y*w y*z - x*w y^2 - x*z]

            sage: M = m.image()
            sage: r = FreeResolution(M, name='S')
            sage: r._m()
            [z^2 - y*w y*z - x*w y^2 - x*z]
        """
        if isinstance(self._module, Ideal_generic):
            return self._module
        if isinstance(self._module, Module_free_ambient):
            return self._module.matrix().transpose()
        if isinstance(self._module, Matrix):
            return self._module.transpose()
        raise ValueError("unable to create a matrix/ideal to build the resolution")


class FiniteFreeResolution_free_module(FiniteFreeResolution):
    r"""
    Free resolutions of a free module.

    INPUT:

    - ``module`` -- a free module or ideal over a PID
    - ``name`` -- the name of the base ring

    EXAMPLES::

        sage: R.<x> = QQ[]
        sage: M = R^3
        sage: v = M([x^2, 2*x^2, 3*x^2])
        sage: w = M([0, x, 2*x])
        sage: S = M.submodule([v, w]); S
        Free module of degree 3 and rank 2 over
         Univariate Polynomial Ring in x over Rational Field
         Echelon basis matrix:
         [  x^2 2*x^2 3*x^2]
         [    0     x   2*x]
        sage: res = S.free_resolution(); res
        S^3 <-- S^2 <-- 0
        sage: ascii_art(res.chain_complex())
                    [  x^2     0]
                    [2*x^2     x]
                    [3*x^2   2*x]
         0 <-- C_0 <-------------- C_1 <-- 0

        sage: R.<x> = PolynomialRing(QQ)
        sage: I = R.ideal([x^4 + 3*x^2 + 2])
        sage: res = I.free_resolution(); res
        S^1 <-- S^1 <-- 0
    """
    @lazy_attribute
    def _maps(self):
        r"""
        Return the maps that define ``self``.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: M = R^3
            sage: v = M([x^2, 2*x^2, 3*x^2])
            sage: w = M([0, x, 2*x])
            sage: S = M.submodule([v, w])
            sage: res = S.free_resolution(); res
            S^3 <-- S^2 <-- 0
            sage: ascii_art(res.chain_complex())
                        [  x^2     0]
                        [2*x^2     x]
                        [3*x^2   2*x]
             0 <-- C_0 <-------------- C_1 <-- 0
            sage: res._maps
            [
            [  x^2     0]
            [2*x^2     x]
            [3*x^2   2*x]
            ]

            sage: R.<x> = PolynomialRing(QQ)
            sage: I = R.ideal([x^4 + 3*x^2 + 2])
            sage: res = I.free_resolution()
            sage: res._maps
            [[x^4 + 3*x^2 + 2]]

            sage: from sage.homology.free_resolution import FreeResolution
            sage: M = matrix([[x^2, 2],
            ....:             [3*x^2, 5],
            ....:             [5*x^2, 4]])
            sage: res = FreeResolution(M.transpose()); res
            S^3 <-- S^2 <-- 0
            sage: res._m()
            [     1      0]
            [   5/2   -x^2]
            [     2 -6*x^2]
            sage: res._maps
            [
            [     1      0]
            [   5/2   -x^2]
            [     2 -6*x^2]
            ]

        An overdetermined system over a PID::

            sage: res = FreeResolution(M); res
            S^2 <-- S^2 <-- 0
            sage: res._m()
            [x^2   0]
            [  2  -1]
            sage: res._maps
            [
            [x^2   0]
            [  2  -1]
            ]
        """
        if isinstance(self._module, Ideal_generic):
            from sage.matrix.constructor import matrix
            return [matrix([[self._module.gen()]])]
        return [self._m()]


class FiniteFreeResolution_singular(FiniteFreeResolution):
    r"""
    Minimal free resolutions of ideals or submodules of free modules
    of multivariate polynomial rings implemented in Singular.

    INPUT:

    - ``module`` -- a submodule of a free module `M` of rank `n` over `S` or
      an ideal of a multi-variate polynomial ring

    - ``name`` -- string (optional); name of the base ring

    - ``algorithm`` -- (default: ``'heuristic'``) Singular algorithm
      to compute a resolution of ``ideal``

    OUTPUT: a minimal free resolution of the ideal

    If ``module`` is an ideal of `S`, it is considered as a submodule of a
    free module of rank `1` over `S`.

    The available algorithms and the corresponding Singular commands
    are shown below:

    ============= ============================
    algorithm     Singular commands
    ============= ============================
    ``minimal``   ``mres(ideal)``
    ``shreyer``   ``minres(sres(std(ideal)))``
    ``standard``  ``minres(nres(std(ideal)))``
    ``heuristic`` ``minres(res(std(ideal)))``
    ============= ============================

    EXAMPLES::

        sage: from sage.homology.free_resolution import FreeResolution
        sage: S.<x,y,z,w> = PolynomialRing(QQ)
        sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
        sage: r = FreeResolution(I); r
        S^1 <-- S^3 <-- S^2 <-- 0
        sage: len(r)
        2

    ::

        sage: FreeResolution(I, algorithm='minimal')
        S^1 <-- S^3 <-- S^2 <-- 0
        sage: FreeResolution(I, algorithm='shreyer')
        S^1 <-- S^3 <-- S^2 <-- 0
        sage: FreeResolution(I, algorithm='standard')
        S^1 <-- S^3 <-- S^2 <-- 0
        sage: FreeResolution(I, algorithm='heuristic')
        S^1 <-- S^3 <-- S^2 <-- 0

    We can also construct a resolution by passing in a matrix defining
    the initial differential::

        sage: m = matrix(S, 1, [z^2 - y*w, y*z - x*w, y^2 - x*z]).transpose()
        sage: r = FreeResolution(m, name='S'); r
        S^1 <-- S^3 <-- S^2 <-- 0
        sage: r.matrix(1)
        [z^2 - y*w y*z - x*w y^2 - x*z]

    An additional construction is using a submodule of a free module::

        sage: M = m.image()
        sage: r = FreeResolution(M, name='S'); r
        S^1 <-- S^3 <-- S^2 <-- 0

    A nonhomogeneous ideal::

        sage: I = S.ideal([z^2 - y*w, y*z - x*w, y^2 - x])
        sage: R = FreeResolution(I); R
        S^1 <-- S^3 <-- S^3 <-- S^1 <-- 0
        sage: R.matrix(2)
        [ y*z - x*w    y^2 - x          0]
        [-z^2 + y*w          0    y^2 - x]
        [         0 -z^2 + y*w -y*z + x*w]
        sage: R.matrix(3)
        [   y^2 - x]
        [-y*z + x*w]
        [ z^2 - y*w]
    """
    def __init__(self, module, name='S', algorithm='heuristic', **kwds) -> None:
        r"""
        Initialize ``self``.

        TESTS::

            sage: from sage.homology.free_resolution import FreeResolution
            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
            sage: r = FreeResolution(I)
            sage: TestSuite(r).run(skip=['_test_pickling'])

            sage: m = matrix(S, 1, [z^2 - y*w, y*z - x*w, y^2 - x*z]).transpose()
            sage: r = FreeResolution(m, name='S')
            sage: TestSuite(r).run(skip=['_test_pickling'])

            sage: M = m.image()
            sage: r = FreeResolution(M, name='S')
            sage: TestSuite(r).run(skip=['_test_pickling'])
        """
        self._algorithm = algorithm
        super().__init__(module, name=name, **kwds)

    @lazy_attribute
    def _maps(self):
        r"""
        Return the maps that define ``self``.

        TESTS::

            sage: from sage.homology.free_resolution import FreeResolution
            sage: S.<x,y,z,w> = PolynomialRing(QQ)
            sage: I = S.ideal([y*w - z^2, -x*w + y*z, x*z - y^2])
            sage: r = FreeResolution(I)
            sage: r._maps
            [
                                             [-y  x]
                                             [ z -y]
            [z^2 - y*w y*z - x*w y^2 - x*z], [-w  z]
            ]
        """
        from sage.libs.singular.singular import si2sa_resolution
        from sage.libs.singular.function import singular_function

        # This ensures the first component of the Singular resolution to be a
        # module, like the later components. This is important when the
        # components are converted to Sage modules.
        module = singular_function("module")
        mod = module(self._m())

        if self._algorithm == 'minimal':
            mres = singular_function('mres')  # syzygy method
            r = mres(mod, 0)
        elif self._algorithm == 'shreyer':
            std = singular_function('std')
            sres = singular_function('sres')  # Shreyer method
            minres = singular_function('minres')
            r = minres(sres(std(mod), 0))
        elif self._algorithm == 'standard':
            nres = singular_function('nres')  # standard basis method
            minres = singular_function('minres')
            r = minres(nres(mod, 0))
        elif self._algorithm == 'heuristic':
            std = singular_function('std')
            res = singular_function('res')    # heuristic method
            minres = singular_function('minres')
            r = minres(res(std(mod), 0))

        return si2sa_resolution(r)
